from VESIcal import activity_models
from VESIcal import calibration_checks
from VESIcal import core
from VESIcal import fugacity_models
from VESIcal import model_classes
from VESIcal import sample_class

import numpy as np
from scipy.optimize import root_scalar
import warnings as w


class water(model_classes.Model):
    """
    Implementation of the Iacono-Marziano et al. (2012) water solubility model, as a Model class.
    Three calibrations are provided- the one incorporating the H2O content as a parameter
    (hydrous), and the one that does not (anhydrous), in addition to the coefficient values given
    incorrect and do not align with the web versions of the model. Which model should be used is
    specified when the methods are called. The default choice is the hydrous model.
    """

    def __init__(self):
        """
        Initialise the model.

        """
        self.set_volatile_species(['H2O'])
        self.set_fugacity_model(fugacity_models.fugacity_idealgas())
        self.set_activity_model(activity_models.activity_idealsolution())
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'temperature', [1000, 1400], calibration_checks.crf_Between, 'oC',
                'IaconoMarzianoWater',
                fail_msg=crmsg_BC_T,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'pressure', [100, 10000], calibration_checks.crf_Between, 'bars',
                'IaconoMarzianoWater',
                fail_msg=crmsg_BC_P,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description)])
        # Not dependent on CO2 conc, H2O dependence dealt with within model.
        self.set_solubility_dependence(False)

        # The oxide masses used in the IM webapp.
        self.IM_oxideMasses = {'Al2O3': 101.96,
                               'CaO':    56.08,
                               'FeO':    71.85,
                               'K2O':    94.2,
                               'MgO':    40.32,
                               'Na2O':   61.98,
                               'SiO2':   60.09,
                               'TiO2':   79.9,
                               'H2O':    18.01}

    def calculate_dissolved_volatiles(self, pressure, temperature, sample, X_fluid=1.0,
                                      coeffs='webapp', **kwargs):
        """
        Calculates the dissolved H2O concentration, using Eq (13) of Iacono-Marziano et al. (2012).
        If using the hydrous parameterization, it will use the scipy.root_scalar routine to find
        the root of the root_dissolved_volatiles method.

        Parameters
        ----------
        pressure    float
            Total pressure in bars.
        temperature     float
            Temperature in C
        sample     pandas Series or dict
            Major element oxides in wt%.
        X_fluid      float
            Mole fraction of H2O in the fluid. Default is 1.0.
        coeffs  str
            Which set of coefficients should be used in the calculations:
            - 'webapp' (default) for the hydrous NBO/O parameterisation coefficients used in
              the Iacono-Marziano webapp.
            - 'manuscript' for the hydrous NBO/O parameterisation coefficients given in the
              Iacono-Marziano et al. (2012) manuscript, but were incorrect (pers.comm.).
            - 'anhydrous' for the anhydrous NBO/O parameterisation coefficients given in the
              Iacono-Marziano et al. (2012) manuscript.

        Returns
        -------
        float
            Dissolved H2O concentration in wt%.
        """

        if coeffs not in ['webapp', 'manuscript', 'anhydrous']:
            raise core.InputError("The coeffs argument must be one of 'webapp', 'manuscript', "
                                  "or 'anhydrous'")

        temperature = temperature + 273.15  # translate T from C to K

        if isinstance(sample, sample_class.Sample) is False:
            raise core.InputError("Sample must be an instance of the Sample class.")
        if pressure < 0:
            raise core.InputError("Pressure must be positive.")
        if X_fluid < 0 or X_fluid > 1:
            raise core.InputError("X_fluid must have a value between 0 and 1.")

        if pressure == 0:
            return 0

        if coeffs == 'webapp' or coeffs == 'manuscript':
            if X_fluid == 0:
                return 0
            H2O = root_scalar(
                self.root_dissolved_volatiles,
                args=(pressure, temperature, sample, X_fluid, coeffs, kwargs),
                x0=1.0, x1=2.0).root
            return H2O
        else:
            a = 0.54
            b = 1.24
            B = -2.95
            C = 0.02

            fugacity = self.fugacity_model.fugacity(pressure=pressure, X_fluid=X_fluid,
                                                    temperature=temperature-273.15, **kwargs)
            if fugacity == 0:
                return 0
            NBO_O = self.NBO_O(sample=sample, coeffs=coeffs)

            H2O = np.exp(a*np.log(fugacity) + b*NBO_O + B + C*pressure/temperature)

            return H2O

    def calculate_equilibrium_fluid_comp(self, pressure, temperature, sample, **kwargs):
        """ Returns 1.0 if a pure H2O fluid is saturated. Returns 0.0 if a pure H2O fluid is
        undersaturated.

        Parameters
        ----------
        pressure     float
            The total pressure of the system in bars.
        temperature     float
            The temperature of the system in C.
        sample         pandas Series or dict
            Major element oxides in wt% (including H2O).

        Returns
        -------
        float
            1.0 if H2O-fluid saturated, 0.0 otherwise.
        """

        if pressure > self.calculate_saturation_pressure(temperature=temperature,
                                                         sample=sample, **kwargs):
            return 0.0
        else:
            return 1.0

    def calculate_saturation_pressure(self, temperature, sample, **kwargs):
        """
        Calculates the pressure at which a pure H2O fluid is saturated, for the given sample
        composition and H2O concentration. Calls the scipy.root_scalar routine, which makes
        repeated called to the calculate_dissolved_volatiles method.

        Parameters
        ----------
        temperature     float
            The temperature of the system in C.
        sample         pandas Series or dict
            Major element oxides in wt% (including H2O).
        X_fluid     float
            The mole fraction of H2O in the fluid. Default is 1.0.

        Returns
        -------
        float
            Calculated saturation pressure in bars.
        """

        if isinstance(sample, sample_class.Sample) is False:
            raise core.InputError("Sample must be an instance of the Sample class.")
        if sample.check_oxide('H2O') is False:
            raise core.InputError("sample must contain H2O.")
        if sample.get_composition('H2O') < 0.0:
            raise core.InputError("Dissolved H2O must be greater than 0 wt%.")

        try:
            # Checks whether the upper bound for the numerical solver returns a positive H2O
            # concentration, and if it doesn't, it will progressively decrease the bound until
            # it does.
            upperbound = 1e5
            while self.calculate_dissolved_volatiles(upperbound, temperature,
                                                     sample, **kwargs) < 0:
                upperbound = upperbound*0.9

            satP = root_scalar(self.root_saturation_pressure, args=(temperature, sample, kwargs),
                               bracket=[1e-15, upperbound]).root
        except Exception:
            w.warn("Saturation pressure not found.", RuntimeWarning, stacklevel=2)
            satP = np.nan
        return satP

    def root_saturation_pressure(self, pressure, temperature, sample, kwargs):
        """ Function called by scipy.root_scalar when finding the saturation pressure using
        calculate_saturation_pressure.

        Parameters
        ----------
        pressure     float
            Pressure guess in bars
        temperature     float
            The temperature of the system in C.
        sample         pandas Series or dict
            Major elements in wt% (normalized to 100%), including H2O.
        kwargs         dictionary
            Additional keyword arguments supplied to calculate_saturation_pressure. Might be
            required for the fugacity or activity models.

        Returns
        -------
        float
            The differece between the dissolved H2O at the pressure guessed, and the H2O
            concentration passed in the sample variable.
        """

        return (sample.get_composition('H2O') -
                self.calculate_dissolved_volatiles(pressure=pressure, temperature=temperature,
                                                   sample=sample, **kwargs))

    def root_dissolved_volatiles(self, h2o, pressure, temperature, sample, X_fluid, coeffs,
                                 kwargs):
        """ Function called by calculate_dissolved_volatiles method when the hydrous
        parameterization is being used.

        Parameters
        ----------
        h2o     float
            Guess for the H2O concentration in wt%.
        pressure     float
            Total pressure in bars.
        temperature     float
            Temperature in K.
        sample         pandas Series or dict
            Major element oxides in wt%.
        X_fluid     float
            Mole fraction of H2O in the fluid.
        coeffs  str
            One of 'webapp','manuscript','anhydrous'.
        kwargs     dictionary
            Keyword arguments

        Returns
        -------
        float
            Difference between H2O guessed and the H2O calculated.
        """

        if coeffs == 'manuscript':
            a = 0.53
            b = 2.35
            B = -3.37
            C = -0.02
        else:
            a = 0.52096846
            b = 2.11575907
            B = -3.24443335
            C = -0.02238884

        sample_copy = sample.change_composition({'H2O': h2o}, inplace=False)

        NBO_O = self.NBO_O(sample=sample_copy, coeffs=coeffs)
        fugacity = self.fugacity_model.fugacity(pressure=pressure, X_fluid=X_fluid,
                                                temperature=temperature, **kwargs)

        return h2o - np.exp(a*np.log(fugacity) + b*NBO_O + B + C*pressure/(temperature+273.15))

    def NBO_O(self, sample, coeffs='webapp'):
        """
        Calculates NBO/O according to Appendix A.1. of Iacono-Marziano et al. (2012). NBO/O
        is calculated on either a hydrous or anhyrous basis, as set when initialising the
        Model class.

        Parameters
        ----------
        sample     pandas Series or dict
            Major element oxides in wt% (including H2O if using the hydrous parameterization).

        coeffs  str
            One of:
            - 'webapp' or 'manuscript' to include H2O in NBO/O
            - 'anhydrous' to exclude H2O from NBO/O

        Returns
        -------
        float
            NBO/O.
        """
        if isinstance(sample, sample_class.Sample) is False:
            raise core.InputError("Sample must be an instance of the Sample class.")
        if all(sample.check_oxide(ox) for ox in ['K2O', 'Na2O', 'CaO', 'MgO',
                                                 'FeO', 'Al2O3', 'SiO2', 'TiO2']) is False:
            raise core.InputError("Sample must contain K2O, Na2O, CaO, MgO, FeO, Al2O3, SiO2, "
                                  "and TiO2.")

        X = sample.get_composition(units='mol_oxides', oxide_masses=self.IM_oxideMasses)

        if 'Fe2O3' in X:
            Fe2O3 = X['Fe2O3']
        else:
            Fe2O3 = 0

        NBO = 2*(X['K2O'] + X['Na2O'] + X['CaO'] + X['MgO'] + X['FeO'] + 2*Fe2O3 - X['Al2O3'])
        Ox = (2*X['SiO2'] + 2*X['TiO2'] + 3*X['Al2O3'] + X['MgO'] + X['FeO'] + 2*Fe2O3 +
              X['CaO'] + X['Na2O'] + X['K2O'])

        if coeffs == 'webapp' or coeffs == 'manuscript':
            if 'H2O' not in X:
                raise core.InputError("sample must contain H2O if using the hydrous"
                                      " parameterization.")
            NBO = NBO + 2*X['H2O']
            Ox = Ox + X['H2O']

        return NBO/Ox


class carbon(model_classes.Model):
    """
    Implementation of the Iacono-Marziano et al. (2012) carbon solubility model, as a Model class.
    """

    def __init__(self):
        """
        Initialise the model.

        """
        self.set_volatile_species(['CO2'])
        self.set_fugacity_model(fugacity_models.fugacity_idealgas())
        self.set_activity_model(activity_models.activity_idealsolution())
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'temperature', [1000, 1400], calibration_checks.crf_Between, 'oC',
                'IaconoMarzianoCarbon',
                fail_msg=crmsg_BC_T,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'pressure', [100, 10000], calibration_checks.crf_Between, 'bars',
                'IaconoMarzianoCarbon',
                fail_msg=crmsg_BC_P,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description)])
        self.set_solubility_dependence(True)

        # The oxide masses used in the IM webapp.
        self.IM_oxideMasses = {'Al2O3': 101.96,
                               'CaO':    56.08,
                               'FeO':    71.85,
                               'K2O':    94.2,
                               'MgO':    40.32,
                               'Na2O':   61.98,
                               'SiO2':   60.09,
                               'TiO2':   79.9,
                               'H2O':    18.01}

    def calculate_dissolved_volatiles(self, pressure, temperature, sample, X_fluid=1,
                                      coeffs='webapp', **kwargs):
        """
        Calculates the dissolved CO2 concentration, using Eq (12) of Iacono-Marziano et al. (2012).
        If using the hydrous parameterization, it will use the scipy.root_scalar routine to find
        the root of the root_dissolved_volatiles method.

        Parameters
        ----------
        pressure    float
            Total pressure in bars.
        temperature     float
            Temperature in C
        sample     Sample class
            Magma major element composition.
        X_fluid      float
            Mole fraction of H2O in the fluid. Default is 1.0.
        coeffs  str
            Which set of coefficients should be used for H2O calculations:
            - 'webapp' (default) for the hydrous NBO/O parameterisation coefficients used in the
              Iacono-Marziano webapp.
            - 'manuscript' for the hydrous NBO/O parameterisation coefficients given in the
              Iacono-Marziano et al. (2012) manuscript.
            - 'anhydrous' for the anhydrous NBO/O parameterisation coefficients given in the
              Iacono-Marziano et al. (2012) manuscript.

        Returns
        -------
        float
            Dissolved H2O concentration in wt%.
        """

        if coeffs not in ['webapp', 'manuscript', 'anhydrous']:
            raise core.InputError("The coeffs argument must be one of 'webapp', 'manuscript', "
                                  "or 'anhydrous'")

        temperature = temperature + 273.15  # translate T from C to K

        if isinstance(sample, sample_class.Sample) is False:
            raise core.InputError("Sample must be an instance of the Sample class.")
        if pressure < 0:
            raise core.InputError("Pressure must be positive.")
        if temperature <= 0:
            raise core.InputError("Temperature must be greater than 0K.")
        if X_fluid < 0 or X_fluid > 1:
            raise core.InputError("X_fluid must have a value between 0 and 1.")

        if pressure == 0:
            return 0

        if coeffs == 'webapp' or coeffs == 'manuscript':
            im_h2o_model = water()
            h2o = im_h2o_model.calculate_dissolved_volatiles(pressure=pressure,
                                                             temperature=temperature-273.15,
                                                             sample=sample, X_fluid=1-X_fluid,
                                                             coeffs=coeffs, **kwargs)

            sample_h2o = sample.change_composition({'H2O': h2o}, inplace=False)

            d = np.array([-16.4, 4.4, -17.1, 22.8])
            a = 1.0
            b = 17.3
            B = -6.0
            C = 0.12

            NBO_O = self.NBO_O(sample=sample_h2o, coeffs=coeffs)

        else:
            im_h2o_model = water()
            h2o = im_h2o_model.calculate_dissolved_volatiles(pressure=pressure,
                                                             temperature=temperature-273.15,
                                                             sample=sample, X_fluid=1-X_fluid,
                                                             coeffs=coeffs, **kwargs)

            sample_h2o = sample.change_composition({'H2O': h2o}, inplace=False)

            d = np.array([2.3, 3.8, -16.3, 20.1])
            a = 1.0
            b = 15.8
            B = -5.3
            C = 0.14

            NBO_O = self.NBO_O(sample=sample, coeffs=coeffs)

        fugacity = self.fugacity_model.fugacity(pressure=pressure, X_fluid=X_fluid,
                                                temperature=temperature-273.15, **kwargs)

        if fugacity == 0:
            return 0

        molarProps = sample_h2o.get_composition(units='mol_oxides',
                                                oxide_masses=self.IM_oxideMasses)

        if all(ox in molarProps for ox in ['Al2O3', 'CaO', 'K2O', 'Na2O',
                                           'FeO', 'MgO', 'Na2O', 'K2O']) is False:
            raise core.InputError("sample must contain Al2O3, CaO, K2O, Na2O, FeO, MgO, Na2O, "
                                  "and K2O.")
        if 'Fe2O3' in molarProps:
            Fe2O3 = molarProps['Fe2O3']
        else:
            Fe2O3 = 0

        x = list()
        if 'H2O' in molarProps:
            x.append(molarProps['H2O'])
        else:
            x.append(0.0)
        x.append(molarProps['Al2O3']/(molarProps['CaO']+molarProps['K2O']+molarProps['Na2O']))
        x.append((molarProps['FeO']+Fe2O3*2+molarProps['MgO']))
        x.append((molarProps['Na2O']+molarProps['K2O']))
        x = np.array(x)

        CO3 = np.exp(np.sum(x*d) + a*np.log(fugacity) + b*NBO_O + B + C*pressure/temperature)
        CO2 = CO3/1e4

        return CO2

    def calculate_equilibrium_fluid_comp(self, pressure, temperature, sample, **kwargs):
        """ Returns 1.0 if a pure CO2 fluid is saturated. Returns 0.0 if a pure CO2 fluid is
        undersaturated.

        Parameters
        ----------
        pressure     float
            The total pressure of the system in bars.
        temperature     float
            The temperature of the system in C.
        sample         Sample class
            Magma major element composition (including H2O).

        Returns
        -------
        float
            1.0 if CO2-fluid saturated, 0.0 otherwise.
        """

        if pressure > self.calculate_saturation_pressure(temperature=temperature, sample=sample,
                                                         **kwargs):
            return 0.0
        else:
            return 1.0

    def calculate_saturation_pressure(self, temperature, sample, **kwargs):
        """
        Calculates the pressure at which a pure CO2 fluid is saturated, for the given sample
        composition and CO2 concentration. Calls the scipy.root_scalar routine, which makes
        repeated called to the calculate_dissolved_volatiles method.

        Parameters
        ----------
        temperature     float
            The temperature of the system in C.
        sample         Sample class
            Magma major element composition (including CO2).

        Returns
        -------
        float
            Calculated saturation pressure in bars.
        """

        if temperature <= 0:
            raise core.InputError("Temperature must be greater than 0K.")
        if isinstance(sample, sample_class.Sample) is False:
            raise core.InputError("Sample must be an instance of the Sample class.")
        if sample.check_oxide('CO2') is False:
            raise core.InputError("sample must contain CO2")
        if sample.get_composition('CO2') < 0:
            raise core.InputError("Dissolved CO2 must be greater than 0 wt%.")

        try:
            satP = root_scalar(self.root_saturation_pressure, args=(temperature, sample, kwargs),
                               bracket=[1e-15, 1e5]).root
        except Exception:
            w.warn("Saturation pressure not found.", RuntimeWarning, stacklevel=2)
            satP = np.nan
        return satP

    def root_saturation_pressure(self, pressure, temperature, sample, kwargs):
        """ Function called by scipy.root_scalar when finding the saturation pressure using
        calculate_saturation_pressure.

        Parameters
        ----------
        pressure     float
            Pressure guess in bars
        temperature     float
            The temperature of the system in C.
        sample       Sample class
            Magma major element composition, including CO2.
        kwargs         dictionary
            Additional keyword arguments supplied to calculate_saturation_pressure. Might be
            required for the fugacity or activity models.

        Returns
        -------
        float
            The differece between the dissolved CO2 at the pressure guessed, and the CO2
            concentration passed in the sample variable.
        """

        return (sample.get_composition('CO2') -
                self.calculate_dissolved_volatiles(pressure=pressure, temperature=temperature,
                                                   sample=sample, **kwargs))

    def NBO_O(self, sample, coeffs='webapp'):
        """
        Calculates NBO/O according to Appendix A.1. of Iacono-Marziano et al. (2012). NBO/O is
        calculated on either a hydrous or anhyrous basis, as set when initialising the Model class.

        Parameters
        ----------
        sample     pandas Series or dict
            Major element oxides in wt% (including H2O if using the hydrous parameterization).

        coeffs  str
            One of:
            - 'webapp' or 'manuscript' to include H2O in NBO/O
            - 'anhydrous' to exclude H2O from NBO/O

        Returns
        -------
        float
            NBO/O.
        """
        if isinstance(sample, sample_class.Sample) is False:
            raise core.InputError("Sample must be an instance of the Sample class.")
        if all(sample.check_oxide(ox) for ox in ['K2O', 'Na2O', 'CaO', 'MgO', 'FeO',
                                                 'Al2O3', 'SiO2', 'TiO2']) is False:
            raise core.InputError("sample must contain K2O, Na2O, CaO, MgO, FeO, Al2O3, SiO2, "
                                  "and TiO2.")

        X = sample.get_composition(units='mol_oxides', oxide_masses=self.IM_oxideMasses)

        if 'Fe2O3' in X:
            Fe2O3 = X['Fe2O3']
        else:
            Fe2O3 = 0

        NBO = 2*(X['K2O']+X['Na2O']+X['CaO']+X['MgO']+X['FeO']+2*Fe2O3-X['Al2O3'])
        Ox = (2*X['SiO2'] + 2*X['TiO2'] + 3*X['Al2O3'] + X['MgO'] + X['FeO'] + 2*Fe2O3 +
              X['CaO'] + X['Na2O'] + X['K2O'])

        if coeffs == 'webapp' or coeffs == 'manuscript':
            if 'H2O' not in X:
                raise core.InputError("sample must contain H2O if using the hydrous "
                                      "parameterization.")
            NBO = NBO + 2*X['H2O']
            Ox = Ox + X['H2O']

        return NBO/Ox


crmsg_BC_T = ("{param_name} ({param_val:.1f} {units}) is outside the broad range suggested by "
              "Iacono-Marziano ({calib_val0:.1f}-{calib_val1:.1f} {units}, although they note "
              "that this model is best calibrated at 1200-1300C). ")
# Warning for Iacono-Marziano pressure
crmsg_BC_P = ("{param_name} ({param_val:.1f} {units}) is outside the broad range suggested by "
              "Iacono-Marziano ({calib_val0:.1f}-{calib_val1:.1f} {units}, although most "
              "calibration experiments were conducted at <5000 bars).  ")


mixed = model_classes.MixedFluid({'H2O': water(), 'CO2': carbon()})



import unittest
import VESIcal as v
import pandas as pd
import pathlib

# Allow unittest to find the file
TEST_FILE = pathlib.Path(__file__).parent.joinpath("ImportTest.xlsx")

class TestImportExcel(unittest.TestCase):
    def assertDataframeEqual(self, a, b, msg):
        """
        Creates a new type of unittest to assert that pd DataFrames
        are equal, inheriting from pandas testing routine
        """
        try:
            pd._testing.assert_frame_equal(a, b)
        except AssertionError as e:
            raise self.failureException(msg) from e

    def setUp(self):
        # Add assertDataframeEqual to unittest test cases
        self.addTypeEqualityFunc(pd.DataFrame, self.assertDataframeEqual)

        self.data_dict = {'SiO2':  [47.95, 47.95],
                     'TiO2':  [1.67, 1.67],
                     'Al2O3': [17.32, 17.32],
                     'FeO':   [10.24, 10.24],
                     'Fe2O3': [0.1, 0.1],
                     'MgO':   [5.76, 5.76],
                     'CaO':   [10.93, 10.93],
                     'Na2O':  [3.45, 3.45],
                     'K2O':   [1.99, 1.99],
                     'P2O5':  [0.51, 0.51],
                     'MnO':   [0.1, 0.1],
                     'H2O':   [2.0, 2.0],
                     'CO2':   [0.1, 0.1],
                     'Cr2O3': [0.0, 0.0],
                     'NiO':   [0.0, 0.0],
                     'CoO':   [0.0, 0.0],
                     'Notes': ['Normal sample', 'Duplicate sample']}
        
        self.df = pd.DataFrame(self.data_dict, 
                               index=['test_samp',
                                      'test_samp-duplicate-1'])

        self.myfile = v.BatchFile(TEST_FILE)

    def test_ImportExcel(self):
        self.assertEqual(self.df, self.myfile.get_data(), 
                         'DataFrame are different')


from VESIcal import activity_models
from VESIcal import calibration_checks
from VESIcal import core
from VESIcal import fugacity_models
from VESIcal import model_classes
from VESIcal import sample_class

import numpy as np
import warnings as w
from scipy.optimize import root_scalar


class carbon(model_classes.Model):
    """ Implementation of the Shishkina et al. (2014) carbon solubility model, as a Model class.
    """
    def __init__(self):
        self.set_volatile_species(['CO2'])
        self.set_fugacity_model(fugacity_models.fugacity_idealgas())
        self.set_activity_model(activity_models.activity_idealsolution())
        self.set_solubility_dependence(False)
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'pressure', [500.0, 5000.0], calibration_checks.crf_Between, 'bar',
                'Shishkina et al. carbon',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'temperature', [1200.0, 1300.0], calibration_checks.crf_Between, 'oC',
                'Shishkina et al. carbon',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'SiO2', [40, 57], calibration_checks.crf_Between, 'wt%',
                'Shishkina et al. carbon',
                fail_msg=crmsg_BC_fail_C1,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description)])

    def PiStar(self, sample):
        """Shishkina et al. (2014) Eq (11)

        Calculates the Pi* parameter for use in calculating CO2 solubility.

        Parameters
        ----------
        sample:        Sample class
            The magma composition stored in a Sample class.

        Returns
        -------
        float
            The value of the Pi* compositional parameter.
        """

        _mols = sample.get_composition(units='mol_cations')

        if all(cation in _mols for cation in ['Ca', 'K', 'Na', 'Mg', 'Fe', 'Si', 'Al']) is False:
            raise core.InputError("To calculate PiStar, values for CaO, K2O, Na2O, MgO, FeO,"
                                  " SiO2, and Al2O3 must be provided in sample.")

        # Calculate assuming all Fe in Fe+2, as done during calibration
        if 'Fe3' in _mols:
            _mols['Fe'] += _mols['Fe3']

        _pi = ((_mols['Ca'] + 0.8*_mols['K'] + 0.7*_mols['Na'] + 0.4*_mols['Mg'] +
               0.4*_mols['Fe'])/(_mols['Si']+_mols['Al']))

        return _pi

    def calculate_dissolved_volatiles(self, pressure, sample, X_fluid=1, **kwargs):
        """ Calculates the dissolved CO2 concentration in wt%, using equation (13) of Shishkina et
        al. (2014).

        Parameters
        ----------
        pressure:    float
            (Total) pressure in bars.
        sample:        Sample class
            Magma composition.
        X_fluid:    float
            The mol-fraction of the fluid that is CO2. Default is 1, i.e. a pure CO2 fluid.

        Returns
        -------
        float
            The dissolved CO2 concentration in wt%.
        """

        if X_fluid < 0 or X_fluid > 1:
            raise core.InputError("X_fluid must have a value between 0 and 1.")
        if pressure < 0:
            raise core.InputError("pressure must be a positive value.")

        PiStar = self.PiStar(sample)
        fugacity = self.fugacity_model.fugacity(pressure=pressure, X_fluid=X_fluid, **kwargs)

        A = 1.150
        B = 6.71
        C = -1.345

        if fugacity == 0:
            return 0
        else:
            return np.exp(A*np.log(fugacity/10)+B*PiStar+C)/1e4

    def calculate_equilibrium_fluid_comp(self, pressure, sample, **kwargs):
        """ Returns 1.0 if a pure CO2 fluid is saturated. Returns 0.0 if a pure CO2 fluid is
        undersaturated.

        Parameters
        ----------
        pressure     float
            The total pressure of the system in bars.
        sample         Sample class
            Magma major element composition.

        Returns
        -------
        float
            1.0 if CO2-fluid saturated, 0.0 otherwise.
        """
        if self.calculate_saturation_pressure(sample=sample, **kwargs) < pressure:
            return 0.0
        else:
            return 1.0

    def calculate_saturation_pressure(self, sample, **kwargs):
        """ Calculates the pressure at which a pure CO2 fluid is saturated, for the given
        sample composition and CO2 concentration. Calls the scipy.root_scalar routine, which makes
        repeated calls to the calculate_dissolved_volatiles method.

        Parameters
        ----------
        sample         Sample class
            Magma major element composition.

        Returns
        -------
        float
            Saturation pressure in bar
        """

        if sample.check_oxide('CO2') is False:
            raise core.InputError("sample must contain CO2.")
        if sample.get_composition('CO2') < 0:
            raise core.InputError("CO2 concentration must be greater than 0 wt%.")

        try:
            satP = root_scalar(self.root_saturation_pressure, bracket=[1e-15, 1e5],
                               args=(sample, kwargs)).root
        except Exception:
            w.warn("Saturation pressure not found.", RuntimeWarning, stacklevel=2)
            satP = np.nan
        return satP

    def root_saturation_pressure(self, pressure, sample, kwargs):
        """ Function called by scipy.root_scalar when finding the saturation pressure using
        calculate_saturation_pressure.

        Parameters
        ----------
        pressure     float
            Pressure guess in bars
        sample         Sample class
            Magma major element composition.
        kwargs         dictionary
            Additional keyword arguments supplied to calculate_saturation_pressure. Might be
            required for the fugacity or activity models.

        Returns
        -------
        float
            The differece between the dissolved CO2 at the pressure guessed, and the CO2
            concentration passed in the sample variable.
        """
        return (self.calculate_dissolved_volatiles(pressure=pressure, sample=sample, **kwargs) -
                sample.get_composition('CO2'))


class water(model_classes.Model):
    """ Implementation of the Shishkina et al. (2014) H2O solubility model as a Model class.
    """
    def __init__(self):
        self.set_volatile_species(['H2O'])
        self.set_fugacity_model(fugacity_models.fugacity_idealgas())
        self.set_activity_model(activity_models.activity_idealsolution())
        self.set_solubility_dependence(False)
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'pressure', [500.0, 5000.0], calibration_checks.crf_Between, 'bar',
                'Shishkina et al. water',
                fail_msg=calibration_checks.crmsg_Between_fail,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'temperature', [1050, 1400], calibration_checks.crf_Between, 'oC',
                'Shishkina et al. water',
                fail_msg=crmsg_BC_fail_T1,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description),
            calibration_checks.CalibrationRange(
                'SiO2', 65, calibration_checks.crf_LessThan, 'wt%',
                'Shishkina et al. water',
                fail_msg=crmsg_LessThan_fail_WaterSi,
                pass_msg=calibration_checks.crmsg_LessThan_pass,
                description_msg=calibration_checks.crmsg_LessThan_description),
            calibration_checks.CalibrationRange(
                'SiO2', 40, calibration_checks.crf_GreaterThan, 'wt%',
                'Shishkina et al. water',
                fail_msg=crmsg_GreaterThan_fail_WaterSi,
                pass_msg=calibration_checks.crmsg_GreaterThan_pass,
                description_msg=calibration_checks.crmsg_GreaterThan_description)])

    def calculate_dissolved_volatiles(self, pressure, sample, X_fluid=1.0, **kwargs):
        """Calculates the dissolved H2O concentration using Eqn (9) of Shishkina et al. (2014).

        Parameters
        ----------
        pressure     float
            Total pressure in bars
        sample         Sample class
            Magma major element composition.
        X_fluid     float
            The mol fraction of H2O in the fluid

        Returns
        -------
        float
            The H2O concentration in wt%
        """
        if isinstance(sample, sample_class.Sample) is False:
            raise core.InputError("Sample must be an instance of the Sample class.")

        if sample.check_oxide('Na2O') is False or sample.check_oxide('K2O') is False:
            raise core.InputError("Na2O and K2O must be present in sample.")

        if pressure < 0:
            raise core.InputError("Pressure must be positive.")

        _mols = sample.get_composition(units='mol_cations')
        _mol_volatiles = 0
        if 'H' in _mols:
            _mol_volatiles += _mols['H']
        if 'C' in _mols:
            _mol_volatiles += _mols['C']

        total_alkalis = (_mols['Na'] + _mols['K'])/(1-_mol_volatiles)

        fugacity = self.fugacity_model.fugacity(pressure, X_fluid=X_fluid, **kwargs)

        a = 3.36e-7 * (fugacity/10)**3 - 2.33e-4*(fugacity/10)**2 + 0.0711*(fugacity/10) - 1.1309
        b = -1.2e-5*(fugacity/10)**2 + 0.0196*(fugacity/10)+1.1297

        return a*total_alkalis + b

    def calculate_equilibrium_fluid_comp(self, pressure, sample, **kwargs):
        """ Returns 1.0 if a pure H2O fluid is saturated.
        Returns 0.0 if a pure H2O fluid is undersaturated.

        Parameters
        ----------
        pressure     float
            The total pressure of the system in bars.
        sample       Sample class
            Magma major element composition (including H2O).

        Returns
        -------
        float
            1.0 if H2O-fluid saturated, 0.0 otherwise.
        """
        if self.calculate_saturation_pressure(sample=sample, **kwargs) < pressure:
            return 0.0
        else:
            return 1.0

    def calculate_saturation_pressure(self, sample, **kwargs):
        """ Calculates the pressure at which a pure H2O fluid is saturated, for the given
        sample composition and H2O concentration. Calls the scipy.root_scalar routine, which makes
        repeated calls to the calculate_dissolved_volatiles method.

        Parameters
        ----------
        sample         Sample class
            Magma major element composition (including H2O).

        Returns
        -------
        float
            Saturation pressure in bar
        """
        if sample.check_oxide('H2O') is False:
            raise core.InputError("sample must contain H2O")
        if sample.get_composition('H2O') < 0:
            raise core.InputError("H2O concentration must be greater than 0 wt%.")

        if sample.get_composition('H2O') < self.calculate_dissolved_volatiles(sample=sample,
                                                                              pressure=0,
                                                                              **kwargs):
            return np.nan

        try:
            satP = root_scalar(self.root_saturation_pressure, bracket=[1e-15, 1e5],
                               args=(sample, kwargs)).root
        except Exception:
            w.warn("Saturation pressure not found.", RuntimeWarning, stacklevel=2)
            satP = np.nan
        return satP

    def root_saturation_pressure(self, pressure, sample, kwargs):
        """ Function called by scipy.root_scalar when finding the saturation pressure using
        calculate_saturation_pressure.

        Parameters
        ----------
        pressure     float
            Pressure guess in bars
        sample         Sample class
            Magma major element composition (including H2O).
        kwargs         dictionary
            Additional keyword arguments supplied to calculate_saturation_pressure. Might be
            required for the fugacity or activity models.

        Returns
        -------
        float
            The differece between the dissolved H2O at the pressure guessed, and the H2O
            concentration passed in the sample variable.
        """
        return (self.calculate_dissolved_volatiles(pressure=pressure, sample=sample, **kwargs) -
                sample.get_composition('H2O'))


crmsg_BC_fail_T1 = ("{param_name} ({param_val:.1f} {units}) is outside the calibration range of "
                    "{model_name} ({calib_val0:.1f}-{calib_val1:.1f} {units}). Note, the authors "
                    "recomend that this model is optimally calibrated between 1150-1250C. ")
crmsg_BC_fail_C1 = ("{param_name} ({param_val:.1f} {units}) is outside the calibration range of "
                    "{model_name} (calculated from the max and minimum concentrations in the "
                    "calibration dataset +-5%; {calib_val0:.1f}-{calib_val1:.1f} {units}).")
# This check warns users if SiO2<40 in the water model
crmsg_GreaterThan_fail_WaterSi = ("{param_name}{units} exceeds the lower calibration limit of "
                                  "{model_name} ({calib_val:.1f} {units}, based on the minimum "
                                  "concentration minus 5% in the calibration dataset). ")
# This check warns users if SiO2>65 wt% , the upper limit suggested by Shishkina et al. (2014)
crmsg_LessThan_fail_WaterSi = ("{param_name} ({param_val:.1f} {units}) exceeds the upper limit of"
                               " {calib_val:.1f} {units} suggested by Shishkina et al. for their "
                               "H2O model. ")


mixed = model_classes.MixedFluid({'H2O': water(), 'CO2': carbon()})

mixed.models[0].calibration_ranges.append(
    calibration_checks.CalibrationRange(
        'X_fluid', 0.0, calibration_checks.crf_MixedFluidWarning, '', 'Shishkina et al. (2014)',
        fail_msg=calibration_checks.crmsg_MixedFluidWarning_fail,
        pass_msg=calibration_checks.crmsg_MixedFluidWarning_pass,
        description_msg=calibration_checks.crmsg_MixedFluidWarning_description))



import pandas as pd
import numpy as np
import warnings as w

from VESIcal import core

from copy import deepcopy, copy


class Sample(object):
    """ The sample class stores compositional information for samples, and contains methods for
    normalization and other compositional calculations.
    """

    def __init__(self, composition, units='wtpt_oxides', default_normalization='none',
                 default_units='wtpt_oxides'):
        """ Initialises the sample class.

        The composition is stored as wtpt. If the composition is provided as wtpt, no
        normalization will be applied. If the composition is supplied as mols, the composition
        will be normalized to 100 wt%.

        Parameters
        ----------
        composition     dict or pandas.Series
            The composition of the sample in the format specified by the composition_type
            parameter. Default is oxides in wtpt.

        units     str
            Specifies the units and type of compositional information passed in the composition
            parameter. Choose from 'wtpt_oxides', 'mol_oxides', 'mol_cations'.

        default_normalization:     None or str
            The type of normalization to apply to the data by default. One of:
                - None (no normalization)
                - 'standard' (default): Normalizes an input composition to 100%.
                - 'fixedvolatiles': Normalizes major element oxides to 100 wt%, including
                  volatiles. The volatile wt% will remain fixed, whilst the other major element
                  oxides are reduced proportionally so that the total is 100 wt%.
                - 'additionalvolatiles': Normalises major element oxide wt% to 100%, assuming it
                  is volatile-free. If H2O or CO2 are passed to the function, their un-normalized
                  values will be retained in addition to the normalized non-volatile oxides,
                  summing to >100%.

        default_units     str
            The type of composition to return by default, one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations
            - mol_singleO
        """

        composition = deepcopy(composition)

        if isinstance(composition, dict):
            composition = pd.Series(composition, dtype='float64')
        elif isinstance(composition, pd.Series) is False:
            raise core.InputError("The composition must be given as either a dictionary or a "
                                  "pandas Series.")

        if units == 'wtpt_oxides':
            self._composition = composition
        elif units == 'mol_oxides':
            self._composition = self._molOxides_to_wtpercentOxides(composition)
        elif units == 'mol_cations':
            self._composition = self._molCations_to_wtpercentOxides(composition)
        else:
            raise core.InputError("Units must be one of 'wtpt_oxides', 'mol_oxides', or "
                                  "'mol_cations'.")

        self.set_default_normalization(default_normalization)
        self.set_default_units(default_units)

    def set_default_normalization(self, default_normalization):
        """ Set the default type of normalization to use with the get_composition() method.

        Parameters
        ----------
        default_normalization:    str
            The type of normalization to apply to the data. One of:
            - 'none' (no normalization)
            - 'standard' (default): Normalizes an input composition to 100%.
            - 'fixedvolatiles': Normalizes major element oxides to 100 wt%, including volatiles.
              The volatile wt% will remain fixed, whilst the other major element oxides are
              reduced proportionally so that the total is 100 wt%.
            - 'additionalvolatiles': Normalises major element oxide wt% to 100%, assuming it is
              volatile-free. If H2O or CO2 are passed to the function, their un-normalized values
              will be retained in addition to the normalized non-volatile oxides, summing to >100%.
        """
        if default_normalization in ['none', 'standard', 'fixedvolatiles', 'additionalvolatiles']:
            self.default_normalization = default_normalization
        else:
            raise core.InputError("The normalization method must be one of 'none', 'standard', "
                                  "'fixedvolatiles', or 'additionalvolatiles'.")

    def set_default_units(self, default_units):
        """ Set the default units of composition to return when using the get_composition() method.

        Parameters
        ----------
        default_units     str
            The type of composition to return, one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations
            - mol_singleO
        """
        if default_units in ['wtpt_oxides', 'mol_oxides', 'mol_cations', 'mol_singleO']:
            self.default_units = default_units
        else:
            raise core.InputError("The units must be one of 'wtpt_oxides', 'mol_oxides', "
                                  "'mol_cations', or 'mol_singleO'.")

    def get_composition(self, species=None, normalization=None, units=None,
                        exclude_volatiles=False, asSampleClass=False, oxide_masses={}):
        """ Returns the composition in the format requested, normalized as requested.

        Parameters
        ----------
        species:    NoneType or str
            The name of the oxide or cation to return the concentration of. If NoneType (default)
            the whole composition will be returned as a pandas.Series. If an oxide is passed, the
            value in wtpt will be returned unless units is set to 'mol_oxides', even if the
            default units for the sample object are mol_oxides. If an element is passed, the
            concentration will be returned as mol_cations, unless 'mol_singleO' is specified as
            units, even if the default units for the sample object are mol_singleO. Unless
            normalization is specified in the method call, none will be applied.

        normalization:     NoneType or str
            The type of normalization to apply to the data. One of:
            - 'none' (no normalization)
            - 'standard' (default): Normalizes an input composition to 100%.
            - 'fixedvolatiles': Normalizes major element oxides to 100 wt%, including volatiles.
              The volatile wt% will remain fixed, whilst the other major element oxides are reduced
              proportionally so that the total is 100 wt%.
            - 'additionalvolatiles': Normalises major element oxide wt% to 100%, assuming it is
              volatile-free. If H2O or CO2 are passed to the function, their un-normalized values
              will be retained in addition to the normalized non-volatile oxides, summing to >100%.
            If NoneType is passed the default normalization option will be used
            (self.default_normalization).

        units:     NoneType or str
            The units of composition to return, one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations
            - mol_singleO
            If NoneType is passed the default units option will be used (self.default_type).

        exclude_volatiles   bool
            If True, volatiles will be excluded from the returned composition, prior to
            normalization and conversion.

        asSampleClass:  bool
            If True, the sample composition will be returned as a sample class, with default
            options. In this case any normalization instructions will be ignored.

        oxide_masses:  dict
            Specify here any oxide masses that should be changed from the VESIcal default. This
            might be useful for recreating other implementations of models that use slightly
            different molecular masses. The default values in VESIcal are given to 3 dp.

        Returns
        -------
        pandas.Series, float, or Sample class
            The sample composition, as specified.
        """
        # Process the oxide_masses variable, if necessary:
        oxideMass = copy(core.oxideMass)
        for ox in oxide_masses:
            if ox in oxideMass:
                oxideMass[ox] = oxide_masses[ox]
            else:
                raise core.InputError("The oxide name provided in oxide_masses is not recognised.")

        # Fetch the default return types if not specified in function call
        if normalization is None and species is None:
            normalization = self.default_normalization
        if units is None and species is None:
            units = self.default_units

        # Check whether to exclude volatiles
        # note that here composition is gotten as wtpt_oxides
        if exclude_volatiles:
            composition = self._composition.copy()
            if 'H2O' in composition.index:
                composition = composition.drop(index='H2O')
            if 'CO2' in composition.index:
                composition = composition.drop(index='CO2')
        else:
            composition = self._composition.copy()

        # Check for a species being provided, if so, work out which units to return.
        if isinstance(species, str):
            if species in composition.index:  # if the requested species has a value, proceed
                if species in core.oxides:
                    if units in ['mol_cations, mol_singleO'] or units is None:
                        units = 'wtpt_oxides'
                elif species in core.cations_to_oxides:
                    if units in ['wtpt_oxides', 'mol_oxides'] or units is None:
                        units = 'mol_cations'
                else:
                    raise core.InputError(species + " was not recognised, check spelling, " +
                                          "capitalization and stoichiometry.")
                if normalization is None:
                    normalization = 'none'
            else:
                return 0.0  # if the requested species has no set value, return a float of 0.0
        elif species is not None:
            raise core.InputError("Species must be either a string or a NoneType.")

        # Get the requested type of composition
        if units == 'wtpt_oxides':
            converted = composition
        elif units == 'mol_oxides':
            converted = self._wtpercentOxides_to_molOxides(composition, oxideMass=oxideMass)
        elif units == 'mol_cations':
            converted = self._wtpercentOxides_to_molCations(composition, oxideMass=oxideMass)
        elif units == 'mol_singleO':
            converted = self._wtpercentOxides_to_molSingleO(composition, oxideMass=oxideMass)
        else:
            raise core.InputError("The units must be one of 'wtpt_oxides', 'mol_oxides', "
                                  "'mol_cations', or 'mol_singleO'.")

        # Do requested normalization
        if normalization == 'none':
            final = converted
        elif normalization == 'standard':
            final = self._normalize_Standard(converted, units=units)
        elif normalization == 'fixedvolatiles':
            final = self._normalize_FixedVolatiles(converted, units=units)
        elif normalization == 'additionalvolatiles':
            final = self._normalize_AdditionalVolatiles(converted, units=units)
        else:
            raise core.InputError("The normalization method must be one of 'none', 'standard', "
                                  "'fixedvolatiles', or 'additionalvolatiles'.")

        if species is None:
            if asSampleClass is False:
                return final
            else:
                return Sample(final)
        elif isinstance(species, str):
            if asSampleClass:
                w.warn("Cannot return single species as Sample class. Returning as float.",
                       RuntimeWarning, stacklevel=2)
            return final[species]

    def change_composition(self, new_composition, units='wtpt_oxides', inplace=True):
        """
        Change the concentration of some component of the composition.

        If the units are moles, they are read as moles relative to the present composition,
        i.e. if you wish to double the moles of MgO, if the present content is 0.1 moles,
        you should provide {'MgO':0.2}. The composition will then be re-normalized. If the
        original composition was provided in un-normalized wt%, the unnormalized total will
        be lost.

        Parameters
        ----------
        new_composition:    dict or pandas.Series
            The components to be updated.
        units:      str
            The units of new_composition. Should be one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations
        inplace:    bool
            If True the object will be modified in place. If False, a copy of the Sample
            object will be created, modified, and then returned.

        Returns
        -------
        Sample class
            Modified Sample class.
        """

        # if new_composition is pandas.Series, convert to dict
        if isinstance(new_composition, pd.Series):
            new_composition = dict(new_composition)

        if inplace is False:
            newsample = deepcopy(self)
            return newsample.change_composition(new_composition, units=units)

        if units == 'wtpt_oxides':
            for ox in new_composition:
                self._composition[ox] = new_composition[ox]

        elif units == 'mol_oxides':
            _comp = self.get_composition(units='mol_oxides')
            for ox in new_composition:
                _comp[ox] = new_composition[ox]
            self._composition = self._molOxides_to_wtpercentOxides(_comp)

        elif units == 'mol_cations':
            _comp = self.get_composition(units='mol_cations')
            for el in new_composition:
                _comp[el] = new_composition[el]
            self._composition = self._molCations_to_wtpercentOxides(_comp)

        else:
            raise core.InputError("Units must be one of 'wtpt_oxides', 'mol_oxides', or "
                                  "'mol_cations'.")

        return self

    def get_formulaweight(self, exclude_volatiles=False):
        """ Converts major element oxides in wt% to the formula weight (on a 1 oxygen basis).

        Parameters
        ----------
        exclude_volatiles   bool
            If True the formula weight will be calculated without volatiles

        Returns
        -------
        float
            The formula weight of the composition, on a one oxygen basis.
        """

        cations = self.get_composition(units='mol_singleO',
                                       exclude_volatiles=exclude_volatiles)

        FW = 15.999
        for cation in cations.index:
            FW += cations[cation]*core.CationMass[core.cations_to_oxides[cation]]

        return FW

    def check_oxide(self, oxide):
        """
        Check whether the sample composition contains the given oxide.

        Parameters
        ----------
        oxide:  str
            Oxide name to check composition for.

        Returns
        -------
        bool
            Whether the composition contains the given oxide, or not.
        """

        if oxide not in core.oxides:
            w.warn("Oxide name not recognised. If it is in your sample, unexpected behaviour "
                   "might occur!",
                   RuntimeWarning, stacklevel=2)
        return oxide in self._composition

    def check_cation(self, cation):
        """
        Check whether the sample composition contains the given cation.

        Parameters
        ----------
        cation:    str
            The element name to check the composition for.

        Returns
        -------
        bool
            Whether the composition contains the given element, or not.
        """

        if cation not in core.cations_to_oxides:
            w.warn("Cation name not recognised. If it is in your sample, unexpected behaviour "
                   "might occur!",
                   RuntimeWarning, stacklevel=2)
        return cation in self.get_composition(units='mol_cations')

    def _normalize_Standard(self, composition, units='wtpt_oxides'):
        """
        Normalizes the given composition to 100 wt%, including volatiles. This method
        is intended only to be called by the get_composition() method.

        Parameters
        ----------
        composition:     pandas.Series
            A rock composition with oxide names as keys and concentrations as values.

        units:      str
            The units of composition. Should be one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations

        Returns
        -------
        pandas.Series
            Normalized oxides in wt%.
        """
        comp = composition.copy()
        comp = dict(comp)

        if units == 'wtpt_oxides':
            normed = pd.Series({k: 100.0 * v / sum(comp.values()) for k, v in comp.items()})
        elif units == 'mol_oxides' or units == 'mol_cations':
            normed = pd.Series({k: v / sum(comp.values()) for k, v in comp.items()})
        else:
            raise core.InputError("Units must be one of 'wtpt_oxides', 'mol_oxides', or "
                                  "'mol_cations'.")

        return normed

    def _normalize_FixedVolatiles(self, composition, units='wtpt_oxides'):
        """
        Normalizes major element oxides to 100 wt%, including volatiles. The volatile wt% will
        remain fixed, whilst the other major element oxides are reduced proportionally so that the
        total is 100 wt%.

        Intended to be called only by the get_composition() method.

        Parameters
        ----------
        composition:     pandas Series
            Major element composition

        units:      str
            The units of composition. Should be one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations

        Returns
        -------
        pandas Series
            Normalized major element oxides.
        """
        comp = composition.copy()
        normalized = pd.Series({}, dtype=float)
        volatiles = 0
        if 'CO2' in list(comp.index):
            volatiles += comp['CO2']
        if 'H2O' in list(comp.index):
            volatiles += comp['H2O']

        for ox in list(comp.index):
            if ox != 'H2O' and ox != 'CO2':
                normalized[ox] = comp[ox]

        if units == 'wtpt_oxides':
            normalized = normalized/np.sum(normalized)*(100-volatiles)
        elif units == 'mol_oxides' or units == 'mol_cations':
            normalized = normalized/np.sum(normalized)*(1-volatiles)
        else:
            raise core.InputError("Units must be one of 'wtpt_oxides', 'mol_oxides', or "
                                  "'mol_cations'.")

        if 'CO2' in list(comp.index):
            normalized['CO2'] = comp['CO2']
        if 'H2O' in list(comp.index):
            normalized['H2O'] = comp['H2O']

        return normalized

    def _normalize_AdditionalVolatiles(self, composition, units='wtpt_oxides'):
        """
        Normalises major element oxide wt% to 100%, assuming it is volatile-free. If H2O or CO2
        are passed to the function, their un-normalized values will be retained in addition to the
        normalized non-volatile oxides, summing to >100%.

        Intended to be called only by the get_composition() method.

        Parameters
        ----------
        composition:     pandas.Series
            Major element composition

        units:      str
            The units of composition. Should be one of:
            - wtpt_oxides (default)
            - mol_oxides
            - mol_cations

        Returns
        -------
        pandas.Series
            Normalized major element oxides.
        """
        comp = composition.copy()
        normalized = pd.Series({}, dtype=float)
        for ox in list(comp.index):
            if ox != 'H2O' and ox != 'CO2':
                normalized[ox] = comp[ox]

        if units == 'wtpt_oxides':
            normalized = normalized/np.sum(normalized)*100
        elif units == 'mol_oxides' or units == 'mol_cations':
            normalized = normalized/np.sum(normalized)
        else:
            raise core.InputError("Units must be one of 'wtpt_oxides', 'mol_oxides', or "
                                  "'mol_cations'.")

        if 'H2O' in comp.index:
            normalized['H2O'] = comp['H2O']
        if 'CO2' in comp.index:
            normalized['CO2'] = comp['CO2']

        return normalized

    def _wtpercentOxides_to_molOxides(self, composition, oxideMass=core.oxideMass):
        """
        Converts a wt% oxide composition to mol oxides, normalised to 1 mol.

        Intended to be called only by the get_composition() method.

        Parameters
        ----------
        composition:    pandas.Series
            Major element oxides in wt%

        oxideMass:  dict
            The molar mass of the oxides. Default is the VESIcal default molar masses.

        Returns
        -------
        pandas.Series
            Molar proportions of major element oxides, normalised to 1.
        """
        molOxides = {}
        comp = composition.copy()
        oxideslist = list(comp.index)

        for ox in oxideslist:
            molOxides[ox] = comp[ox]/oxideMass[ox]

        molOxides = pd.Series(molOxides)
        molOxides = molOxides/molOxides.sum()

        return molOxides

    def _wtpercentOxides_to_molCations(self, composition, oxideMass=core.oxideMass):
        """
        Converts a wt% oxide composition to molar proportions of cations (normalised to 1).

        Intended to be called only by the get_composition() method.

        Parameters
        ----------
        composition        pandas.Series
            Major element oxides in wt%.

        oxideMass:  dict
            The molar mass of the oxides. Default is the VESIcal default molar masses.

        Returns
        -------
        pandas.Series
            Molar proportions of cations, normalised to 1.
        """
        molCations = {}
        comp = composition.copy()
        oxideslist = list(comp.index)

        for ox in oxideslist:
            cation = core.oxides_to_cations[ox]
            molCations[cation] = core.CationNum[ox]*comp[ox]/oxideMass[ox]

        molCations = pd.Series(molCations)
        molCations = molCations/molCations.sum()

        return molCations

    def _wtpercentOxides_to_molSingleO(self, composition, oxideMass=core.oxideMass):
        """
        Constructs the chemical formula, on a single oxygen basis, from wt% oxides.

        Intended to be called only by the get_composition() method.

        Parameters
        ----------
        composition        pandas.Series
            Major element oxides in wt%

        oxideMass:  dict
            The molar mass of the oxides. Default is the VESIcal default molar masses.

        Returns
        -------
        pandas.Series
            The chemical formula of the composition, on a single oxygen basis. Each element is
            a separate entry in the Series.
        """
        molCations = {}
        comp = composition.copy()

        oxideslist = list(comp.index)

        total_O = 0.0
        for ox in oxideslist:
            cation = core.oxides_to_cations[ox]
            molCations[cation] = core.CationNum[ox]*comp[ox]/oxideMass[ox]
            total_O += core.OxygenNum[ox]*comp[ox]/oxideMass[ox]

        molCations = pd.Series(molCations)
        molCations = molCations/total_O

        return molCations

    def _molOxides_to_wtpercentOxides(self, composition, oxideMass=core.oxideMass):
        """
        Converts mol oxides to wt% oxides. Returned composition is normalized to 100 wt%.

        Parameters
        ----------
        composition:     pandas.Series
            mol fraction oxides

        oxideMass:  dict
            The molar mass of the oxides. Default is the VESIcal default molar masses.

        Returns
        -------
        pandas.Series
            wt% oxides normalized to 100 wt%.
        """

        comp = composition.copy()
        wtpt = {}

        for ox in composition.index:
            wtpt[ox] = comp[ox]*oxideMass[ox]

        wtpt = pd.Series(wtpt)
        wtpt = wtpt/wtpt.sum()*100

        return wtpt

    def _molOxides_to_molCations(self, composition):
        """
        Converts mol oxides to mol cations. Returned composition is normalized to 1 mol
        cations.

        Parameters
        ----------
        composition:     pandas.Series
            mole fraction oxides

        Returns
        -------
        pandas.Series
            mole fraction cations
        """

        comp = composition.copy()
        molcations = {}

        for ox in comp.index:
            molcations[core.oxides_to_cations[ox]] = comp[ox]*core.CationNum[ox]

        molcations = pd.Series(molcations)
        molcations = molcations/molcations.sum()

        return molcations

    def _molCations_to_wtpercentOxides(self, composition, oxideMass=core.oxideMass):
        """
        Converts mole fraction cations to wt% oxides, normalized to 100 wt%.

        Parameters
        ----------
        composition:     pandas.Series
            Mole fraction cations

        oxideMass:  dict
            The molar mass of the oxides. Default is the VESIcal default molar masses.

        Returns
        -------
        pandas.Series
            Wt% oxides, normalized to 100 wt%.
        """

        comp = composition.copy()
        wtpt = {}

        for el in comp.index:
            wtpt[core.cations_to_oxides[el]] = (
                comp[el] / core.CationNum[core.cations_to_oxides[el]] *
                oxideMass[core.cations_to_oxides[el]])

        wtpt = pd.Series(wtpt)
        wtpt = wtpt/wtpt.sum()*100

        return wtpt

    def _molCations_to_molOxides(self, composition):
        """
        Converts mole fraction cations to mole fraction oxides, normalized to 1 mole.

        Parameters
        ----------
        composition:     pandas.Series
            Mole fraction cations

        Returns
        -------
        pandas.Series
            Mole fraction oxides, normalized to one.
        """
        comp = composition.copy()
        moloxides = {}

        for el in comp.index:
            moloxides[core.cations_to_oxides[el]] = (
                comp[el]/core.CationNum[core.cations_to_oxides[el]])

        moloxides = pd.Series(moloxides)
        moloxides = moloxides/moloxides.sum()

        return moloxides



import unittest
import VESIcal as v

class TestDissolvedVolatiles(unittest.TestCase):
    def setUp(self):
        # Sample with units as wtpt_oxides
        self.majors_wtpt = {'SiO2':    47.95,
                         'TiO2':    1.67,
                         'Al2O3':   17.32,
                         'FeO':     10.24,
                         'Fe2O3':   0.1,
                         'MgO':     5.76,
                         'CaO':     10.93,
                         'Na2O':    3.45,
                         'K2O':     1.99,
                         'P2O5':    0.51,
                         'MnO':     0.1}

        # create Sample object and set default units to wtpt_oxides
        self.sample_wtpt = v.Sample(self.majors_wtpt)
        self.sample_wtpt.set_default_units("wtpt_oxides")

        # create sample and set default units to mol_oxides
        self.sample_molox = v.Sample(self.majors_wtpt)
        self.sample_molox.set_default_units("mol_oxides")

        # BatchFile with test sample as defined above in wtpt_oxides
        try:
            self.batch_wtpt = v.BatchFile('BatchTest.xlsx', units='wtpt_oxides')
        except:
            self.batch_wtpt = v.BatchFile('tests/BatchTest.xlsx', units='wtpt_oxides')
        self.batch_wtpt.set_default_units("wtpt_oxides")

        # BatchFile with test sample as defined above in mol_oxides
        try:
            self.batch_molox = v.BatchFile('BatchTest.xlsx')
        except:
            self.batch_molox = v.BatchFile('tests/BatchTest.xlsx')
        self.batch_molox.set_default_units("mol_oxides")

        # solubilities calculated with VESIcal in wt% and translated externally to mol fraction
        self.shishkinaMixed_wtpt =      {'H2O_liq': 2.24019,
                                        'CO2_liq': 0.028104173}
        self.shishkinaMixed_molox =     {'H2O_liq': 0.07415139,
                                        'CO2_liq': 0.000379788}

        self.dixonMixed_wtpt =          {'H2O_liq': 2.06748958,
                                        'CO2_liq': 0.035434893}
        self.dixonMixed_molox =         {'H2O_liq': 0.0688214,
                                        'CO2_liq': 0.000481216}

        self.iaconomarzianoMixed_wtpt = {'H2O_liq': 2.0853469,
                                        'CO2_liq': 0.0500067}
        self.iaconomarzianoMixed_molox ={'H2O_liq': 0.0693591,
                                        'CO2_liq': 0.000680900}

        self.liuMixed_wtpt =            {'H2O_liq': 2.293923234,
                                        'CO2_liq': 0.029485739}
        self.liuMixed_molox =           {'H2O_liq': 0.075793677,
                                        'CO2_liq': 0.000398905}

        self.magmasat_wtpt =            {'H2O_liq': 1.982045999,
                                        'CO2_liq': 0.044939669}
        self.magmasat_molox =           {'H2O_liq': 0.06615768,
                                        'CO2_liq': 0.000614167}

        self.allisonCarbon_wtpt             = 0.06101661
        self.allisonCarbon_molox            = 0.00089276
        self.allisonCarbon_sunset_wtpt      = 0.03249080
        self.allisonCarbon_sunset_molox     = 0.00047558
        self.allisonCarbon_sfvf_wtpt        = 0.02660018
        self.allisonCarbon_sfvf_molox       = 0.00038939
        self.allisonCarbon_erebus_wtpt      = 0.03314730
        self.allisonCarbon_erebus_molox     = 0.00048519
        self.allisonCarbon_vesuvius_wtpt    = 0.06101661
        self.allisonCarbon_vesuvius_molox   = 0.00089276
        self.allisonCarbon_etna_wtpt        = 0.04799322
        self.allisonCarbon_etna_molox       = 0.00070234
        self.allisonCarbon_stromboli_wtpt   = 0.03216744
        self.allisonCarbon_stromboli_molox  = 0.00047085
        self.mooreWater_wtpt                = 2.24343860
        self.mooreWater_molox               = 0.07427734

        self.mixed_dict =   {"ShishkinaIdealMixing": {'wtpt_oxides': self.shishkinaMixed_wtpt, 'mol_oxides': self.shishkinaMixed_molox},
                             "Dixon"               : {'wtpt_oxides': self.dixonMixed_wtpt, 'mol_oxides': self.dixonMixed_molox},
                             "IaconoMarziano"      : {'wtpt_oxides': self.iaconomarzianoMixed_wtpt, 'mol_oxides': self.iaconomarzianoMixed_molox},
                             "Liu"                 : {'wtpt_oxides': self.liuMixed_wtpt, 'mol_oxides': self.liuMixed_molox},
                             "MagmaSat"            : {'wtpt_oxides': self.magmasat_wtpt, 'mol_oxides': self.magmasat_molox}
                             }

        self.nonmixed_dict = {"AllisonCarbon"           : {'wtpt_oxides': self.allisonCarbon_wtpt, 'mol_oxides': self.allisonCarbon_molox},
                              "AllisonCarbon_sunset"    : {'wtpt_oxides': self.allisonCarbon_sunset_wtpt, 'mol_oxides': self.allisonCarbon_sunset_molox},
                              'AllisonCarbon_sfvf'      : {'wtpt_oxides': self.allisonCarbon_sfvf_wtpt, 'mol_oxides': self.allisonCarbon_sfvf_molox},
                              'AllisonCarbon_erebus'    : {'wtpt_oxides': self.allisonCarbon_erebus_wtpt, 'mol_oxides': self.allisonCarbon_erebus_molox},
                              'AllisonCarbon_vesuvius'  : {'wtpt_oxides': self.allisonCarbon_vesuvius_wtpt, 'mol_oxides': self.allisonCarbon_vesuvius_molox},
                              'AllisonCarbon_etna'      : {'wtpt_oxides': self.allisonCarbon_etna_wtpt, 'mol_oxides': self.allisonCarbon_etna_molox},
                              'AllisonCarbon_stromboli' : {'wtpt_oxides': self.allisonCarbon_stromboli_wtpt, 'mol_oxides': self.allisonCarbon_stromboli_molox},
                              'MooreWater'              : {'wtpt_oxides': self.mooreWater_wtpt, 'mol_oxides': self.mooreWater_molox},
                            }

    def test_calculate_single_wtpt_mixed(self):
        for model in self.mixed_dict.keys():
            calcd_result = v.calculate_dissolved_volatiles(self.sample_wtpt, pressure=1000, temperature=1000, X_fluid=0.5, model=model, verbose=False).result
            known_result = self.mixed_dict[model]['wtpt_oxides']
            for k in calcd_result.keys():
                self.assertAlmostEqual(calcd_result[k], known_result[k], places=4)

    # def test_calculate_batch_wtpt_mixed(self):
    #     for model in self.mixed_dict.keys():
    #         calcd_result = self.batch_wtpt.calculate_dissolved_volatiles(self.sample_wtpt, pressure=1000, temperature=1000, X_fluid=0.5, model=model, verbose=False).result
    #         known_result = self.mixed_dict[model]['wtpt_oxides']
    #         for k in calcd_result.keys():
    #             self.assertAlmostEqual(calcd_result[k], known_result[k], places=4)

    def test_calculate_single__wtpt_nonmixed(self):
        for model in self.nonmixed_dict.keys():
            calcd_result = v.calculate_dissolved_volatiles(self.sample_wtpt, pressure=1000, temperature=1000, X_fluid=0.5, model=model, verbose=False).result
            known_result = self.nonmixed_dict[model]['wtpt_oxides']
            self.assertAlmostEqual(calcd_result, known_result, places=4)

    def test_calculate_single__molox_mixed(self):
        for model in self.mixed_dict.keys():
            calcd_result = v.calculate_dissolved_volatiles(self.sample_molox, pressure=1000, temperature=1000, X_fluid=0.5, model=model, verbose=False).result
            known_result = self.mixed_dict[model]['mol_oxides']
            for k in calcd_result.keys():
                self.assertAlmostEqual(calcd_result[k], known_result[k], places=4)

    def test_calculate_single__molox_nonmixed(self):
        for model in self.nonmixed_dict.keys():
            calcd_result = v.calculate_dissolved_volatiles(self.sample_molox, pressure=1000, temperature=1000, X_fluid=0.5, model=model, verbose=False).result
            known_result = self.nonmixed_dict[model]['mol_oxides']
            self.assertAlmostEqual(calcd_result, known_result, places=4)

class TestSaturationPressure(unittest.TestCase):
    def setUp(self):
        # Sample with units as wtpt_oxides
        self.majors_wtpt = {'SiO2':    47.95,
                         'TiO2':    1.67,
                         'Al2O3':   17.32,
                         'FeO':     10.24,
                         'Fe2O3':   0.1,
                         'MgO':     5.76,
                         'CaO':     10.93,
                         'Na2O':    3.45,
                         'K2O':     1.99,
                         'P2O5':    0.51,
                         'MnO':     0.1,
                         'H2O':     2.0,
                         'CO2':     0.1}

        # Set conditions of calculation
        self.temperature = 1000

        # create Sample object and set default units to wtpt_oxides
        self.sample_wtpt = v.Sample(self.majors_wtpt)
        self.sample_wtpt.set_default_units("wtpt_oxides")

        # create sample and set default units to mol_oxides
        self.sample_molox = v.Sample(self.majors_wtpt)
        self.sample_molox.set_default_units("mol_oxides")

        # BatchFile with test sample as defined above in wtpt_oxides
        try:
            self.batch_wtpt = v.BatchFile('BatchTest.xlsx', units='wtpt_oxides')
        except:
            self.batch_wtpt = v.BatchFile('tests/BatchTest.xlsx', units='wtpt_oxides')
        self.batch_wtpt.set_default_units("wtpt_oxides")

        # BatchFile with test sample as defined above in mol_oxides
        try:
            self.batch_molox = v.BatchFile('BatchTest.xlsx')
        except:
            self.batch_molox = v.BatchFile('tests/BatchTest.xlsx')
        self.batch_molox.set_default_units("mol_oxides")

        # saturation pressures calculated with VESIcal
        self.shishkinaMixed =      1903.1192979127836
        self.dixonMixed =          1847.1637265676327
        self.iaconomarzianoMixed = 1437.268470833670
        self.liuMixed =            2157.191497153953
        self.magmasat =            1630  # Generated with MagmaSat app

        self.shishkinaCarbon           = 1507.647195272281
        self.dixonCarbon               = 1375.61109469857
        self.iaconomarzianoCarbon      = 1215.0221712464863
        self.allisonCarbon             = 821.9428189121331
        self.allisonCarbon_sunset      = 1468.1852834512158
        self.allisonCarbon_sfvf        = 1728.744343059253
        self.allisonCarbon_erebus      = 1439.8801634549727
        self.allisonCarbon_vesuvius    = 821.9428189121331
        self.allisonCarbon_etna        = 1039.5129481979354
        self.allisonCarbon_stromboli   = 1472.5760950857439
        self.liuCarbon                 = 2246.2067748765

        self.shishkinaWater            = 395.4721026237296
        self.dixonWater                = 431.1140172567279
        self.iaconomarzianoWater       = 459.7132843026062
        self.mooreWater                = 366.7939178950552
        self.liuWater                  = 374.4357814145504

        self.mixed_dict =   {"ShishkinaIdealMixing": self.shishkinaMixed,
                             "Dixon"               : self.dixonMixed,
                             "IaconoMarziano"      : self.iaconomarzianoMixed,
                             "Liu"                 : self.liuMixed,
                             "MagmaSat"            : self.magmasat
                             }

        self.carbon_dict = {"ShishkinaCarbon"        : self.shishkinaCarbon,
                           "DixonCarbon"             : self.dixonCarbon,
                           "IaconoMarzianoCarbon"    : self.iaconomarzianoCarbon,
                           "AllisonCarbon"           : self.allisonCarbon,
                           "AllisonCarbon_sunset"    : self.allisonCarbon_sunset,
                           'AllisonCarbon_sfvf'      : self.allisonCarbon_sfvf,
                           'AllisonCarbon_erebus'    : self.allisonCarbon_erebus,
                           'AllisonCarbon_vesuvius'  : self.allisonCarbon_vesuvius,
                           'AllisonCarbon_etna'      : self.allisonCarbon_etna,
                           'AllisonCarbon_stromboli' : self.allisonCarbon_stromboli,
                           'LiuCarbon'               : self.liuCarbon
                           }

        self.water_dict = {"ShishkinaWater"       : self.shishkinaWater,
                          "DixonWater"            : self.dixonWater,
                          "IaconoMarzianoWater"   : self.iaconomarzianoWater,
                          "MooreWater"            : self.mooreWater,
                          "LiuWater"              : self.liuWater
                          }

    def test_calculate_single_wtpt_mixed(self):
        for model in self.mixed_dict.keys():
            calcd_result = v.calculate_saturation_pressure(self.sample_wtpt, temperature=self.temperature, model=model, verbose=False).result
            known_result = self.mixed_dict[model]
            self.assertAlmostEqual(calcd_result, known_result, places=4)

    def test_calculate_batch_wtpt_mixed(self):
        for model in self.mixed_dict.keys():
            batch_result = self.batch_wtpt.calculate_saturation_pressure(temperature=self.temperature, model=model, verbose=True)
            calcd_result = batch_result['SaturationP_bars_VESIcal'].loc['test_samp']
            known_result = self.mixed_dict[model]
            #print("Pi* wtpt = " + str(batch_result['PiStar_VESIcal'].loc['test_samp']))
            self.assertAlmostEqual(calcd_result, known_result, places=4)

    def test_calculate_single_wtpt_carbon(self):
        for model in self.carbon_dict.keys():
            calcd_result = v.calculate_saturation_pressure(self.sample_wtpt, temperature=self.temperature, model=model, verbose=False).result
            known_result = self.carbon_dict[model]
            self.assertAlmostEqual(calcd_result, known_result, places=4)

    def test_calculate_batch_wtpt_carbon(self):
        for model in self.carbon_dict.keys():
            batch_result = self.batch_wtpt.calculate_saturation_pressure(temperature=self.temperature, model=model, verbose=False)
            calcd_result = batch_result['SaturationP_bars_VESIcal'].loc['test_samp']
            known_result = self.carbon_dict[model]
            self.assertAlmostEqual(calcd_result, known_result, places=4)

    def test_calculate_single_wtpt_water(self):
        for model in self.water_dict.keys():
            calcd_result = v.calculate_saturation_pressure(self.sample_wtpt, temperature=self.temperature, model=model, verbose=False).result
            known_result = self.water_dict[model]
            self.assertAlmostEqual(calcd_result, known_result, places=4)

    def test_calculate_batch_wtpt_water(self):
        for model in self.water_dict.keys():
            batch_result = self.batch_wtpt.calculate_saturation_pressure(temperature=self.temperature, model=model, verbose=False)
            calcd_result = batch_result['SaturationP_bars_VESIcal'].loc['test_samp']
            known_result = self.water_dict[model]
            self.assertAlmostEqual(calcd_result, known_result, places=4)

    def test_calculation_single_molox_mixed(self):
        for model in self.mixed_dict.keys():
            calcd_result = v.calculate_saturation_pressure(self.sample_molox, temperature=self.temperature, model=model, verbose=False).result
            known_result = self.mixed_dict[model]
            self.assertAlmostEqual(calcd_result, known_result, places=4)

    def test_calculation_batch_molox_mixed(self):
        for model in self.mixed_dict.keys():
            batch_result = self.batch_molox.calculate_saturation_pressure(temperature=self.temperature, model=model, verbose=True)
            calcd_result = batch_result['SaturationP_bars_VESIcal'].loc['test_samp']
            known_result = self.mixed_dict[model]
            #print("Pi* molox = " + str(batch_result['PiStar_VESIcal'].loc['test_samp']))
            self.assertAlmostEqual(calcd_result, known_result, places=4)

    def test_calculate_single_molox_carbon(self):
        for model in self.carbon_dict.keys():
            calcd_result = v.calculate_saturation_pressure(self.sample_molox, temperature=self.temperature, model=model, verbose=False).result
            known_result = self.carbon_dict[model]
            self.assertAlmostEqual(calcd_result, known_result, places=4)

    def test_calculate_batch_molox_carbon(self):
        for model in self.carbon_dict.keys():
            batch_result = self.batch_molox.calculate_saturation_pressure(temperature=self.temperature, model=model, verbose=False)
            calcd_result = batch_result['SaturationP_bars_VESIcal'].loc['test_samp']
            known_result = self.carbon_dict[model]
            self.assertAlmostEqual(calcd_result, known_result, places=4)

    def test_calculate_single_molox_water(self):
        for model in self.water_dict.keys():
            calcd_result = v.calculate_saturation_pressure(self.sample_molox, temperature=self.temperature, model=model, verbose=False).result
            known_result = self.water_dict[model]
            self.assertAlmostEqual(calcd_result, known_result, places=4)

    def test_calculate_batch_molox_water(self):
        for model in self.water_dict.keys():
            batch_result = self.batch_molox.calculate_saturation_pressure(temperature=self.temperature, model=model, verbose=False)
            calcd_result = batch_result['SaturationP_bars_VESIcal'].loc['test_samp']
            known_result = self.water_dict[model]
            self.assertAlmostEqual(calcd_result, known_result, places=4)

class TestEquilibriumFluidComp(unittest.TestCase):
    def setUp(self):
        # Sample with units as wtpt_oxides
        self.majors_wtpt = {'SiO2':    47.95,
                         'TiO2':    1.67,
                         'Al2O3':   17.32,
                         'FeO':     10.24,
                         'Fe2O3':   0.1,
                         'MgO':     5.76,
                         'CaO':     10.93,
                         'Na2O':    3.45,
                         'K2O':     1.99,
                         'P2O5':    0.51,
                         'MnO':     0.1,
                         'H2O':     2.0,
                         'CO2':     0.1}

        # Set conditions of calculation
        self.temperature = 1000
        self.pressure = 500

        # create Sample object and set default units to wtpt_oxides
        self.sample_wtpt = v.Sample(self.majors_wtpt)
        self.sample_wtpt.set_default_units("wtpt_oxides")

        # create sample and set default units to mol_oxides
        self.sample_molox = v.Sample(self.majors_wtpt)
        self.sample_molox.set_default_units("mol_oxides")

        # equilibrium fluid comps calculated with VESIcal
        self.shishkinaMixed =      {'H2O': 0.7158408854774682, 'CO2': 0.28415911452253184}
        self.dixonMixed =          {'H2O': 0.7750859842655139, 'CO2': 0.2249140157344861}
        self.iaconomarzianoMixed = {'H2O': 0.8103088488568084, 'CO2': 0.18969115114319157}
        self.liuMixed =            {'H2O': 0.7066707740811572, 'CO2': 0.2933292259188428}
        self.magmasat =            {'CO2': 0.219354223233457, 'H2O': 0.780645776766543}

        self.shishkinaCarbon           = 1.0
        self.dixonCarbon               = 1.0
        self.iaconomarzianoCarbon      = 1.0
        self.allisonCarbon             = 1.0
        self.allisonCarbon_sunset      = 1.0
        self.allisonCarbon_sfvf        = 1.0
        self.allisonCarbon_erebus      = 1.0
        self.allisonCarbon_vesuvius    = 1.0
        self.allisonCarbon_etna        = 1.0
        self.allisonCarbon_stromboli   = 1.0
        self.liuCarbon                 = 0

        self.shishkinaWater            = 0.0
        self.dixonWater                = 0.0
        self.iaconomarzianoWater       = 0.0
        self.mooreWater                = 0.7002733835379757
        self.liuWater                  = 0.758191771357082

        self.mixed_dict =   {"ShishkinaIdealMixing": self.shishkinaMixed,
                             "Dixon"               : self.dixonMixed,
                             "IaconoMarziano"      : self.iaconomarzianoMixed,
                             "Liu"                 : self.liuMixed,
                             "MagmaSat"            : self.magmasat
                             }

        self.carbon_dict = {"ShishkinaCarbon"        : self.shishkinaCarbon,
                           "DixonCarbon"             : self.dixonCarbon,
                           "IaconoMarzianoCarbon"    : self.iaconomarzianoCarbon,
                           "AllisonCarbon"           : self.allisonCarbon,
                           "AllisonCarbon_sunset"    : self.allisonCarbon_sunset,
                           'AllisonCarbon_sfvf'      : self.allisonCarbon_sfvf,
                           'AllisonCarbon_erebus'    : self.allisonCarbon_erebus,
                           'AllisonCarbon_vesuvius'  : self.allisonCarbon_vesuvius,
                           'AllisonCarbon_etna'      : self.allisonCarbon_etna,
                           'AllisonCarbon_stromboli' : self.allisonCarbon_stromboli,
                           'LiuCarbon'               : self.liuCarbon
                           }

        self.water_dict = {"ShishkinaWater"       : self.shishkinaWater,
                          "DixonWater"            : self.dixonWater,
                          "IaconoMarzianoWater"   : self.iaconomarzianoWater,
                          "MooreWater"            : self.mooreWater,
                          "LiuWater"              : self.liuWater
                          }

    def test_calculate_wtpt_mixed(self):
        for model in self.mixed_dict.keys():
            calcd_result = v.calculate_equilibrium_fluid_comp(self.sample_wtpt,
                                                              temperature=self.temperature,
                                                              pressure=self.pressure,
                                                              model=model,
                                                              verbose=False).result
            known_result = self.mixed_dict[model]
            for key in known_result.keys():
                self.assertAlmostEqual(calcd_result[key], known_result[key], places=4)

    def test_calculate_wtpt_carbon(self):
        for model in self.carbon_dict.keys():
            calcd_result = v.calculate_equilibrium_fluid_comp(self.sample_wtpt,
                                                              temperature=self.temperature,
                                                              pressure=self.pressure,
                                                              model=model,
                                                              verbose=False).result
            known_result = self.carbon_dict[model]
            self.assertAlmostEqual(calcd_result, known_result, places=4)

    def test_calculate_wtpt_water(self):
        for model in self.water_dict.keys():
            calcd_result = v.calculate_equilibrium_fluid_comp(self.sample_wtpt,
                                                              temperature=self.temperature,
                                                              pressure=self.pressure,
                                                              model=model,
                                                              verbose=False).result
            known_result = self.water_dict[model]
            self.assertAlmostEqual(calcd_result, known_result, places=4)

    def test_calculation_molox_mixed(self):
        for model in self.mixed_dict.keys():
            calcd_result = v.calculate_equilibrium_fluid_comp(self.sample_molox,
                                                              temperature=self.temperature,
                                                              pressure=self.pressure,
                                                              model=model,
                                                              verbose=False).result
            known_result = self.mixed_dict[model]
            for key in known_result.keys():
                self.assertAlmostEqual(calcd_result[key], known_result[key], places=4)

    def test_calculate_molox_carbon(self):
        for model in self.carbon_dict.keys():
            calcd_result = v.calculate_equilibrium_fluid_comp(self.sample_molox,
                                                              temperature=self.temperature,
                                                              pressure=self.pressure,
                                                              model=model,
                                                              verbose=False).result
            known_result = self.carbon_dict[model]
            self.assertAlmostEqual(calcd_result, known_result, places=4)

    def test_calculate_molox_water(self):
        for model in self.water_dict.keys():
            calcd_result = v.calculate_equilibrium_fluid_comp(self.sample_molox,
                                                              temperature=self.temperature,
                                                              pressure=self.pressure,
                                                              model=model,
                                                              verbose=False).result
            known_result = self.water_dict[model]
            self.assertAlmostEqual(calcd_result, known_result, places=4)

if __name__ == '__main__':
    unittest.main()



from VESIcal import activity_models
from VESIcal import calibration_checks
from VESIcal import core
from VESIcal import fugacity_models
from VESIcal import model_classes
from VESIcal import sample_class

import numpy as np
from scipy.optimize import root_scalar
import warnings as w


class carbon(model_classes.Model):
    """
    Implementation of the Dixon (1997) carbon solubility model, as a Model class.
    """

    def __init__(self):
        self.set_volatile_species(['CO2'])
        self.set_fugacity_model(fugacity_models.fugacity_MRK_co2())
        self.set_activity_model(activity_models.activity_idealsolution())
        self.set_calibration_ranges([])
        self.set_solubility_dependence(False)

    def calculate_dissolved_volatiles(self, pressure, sample, X_fluid=1.0, **kwargs):
        """Calculates the dissolved CO2 concentration using Eqn (3) of Dixon (1997).

        Parameters
        ----------
        pressure  float
            Total pressure in bars.
        sample      Sample class
            Magma major element composition.
        X_fluid      float
            The mol fraction of CO2 in the fluid.

        Returns
        -------
        float
            The CO2 concentration in wt%.
        """

        if X_fluid < 0 or X_fluid > 1:
            raise core.InputError("X_fluid must have a value between 0 and 1.")
        if pressure < 0:
            raise core.InputError("Pressure must be positive.")
        if sample.check_oxide('SiO2') is False:
            raise core.InputError("sample must contain SiO2.")

        if pressure == 0:
            return 0

        XCO3 = self.molfrac_molecular(pressure=pressure, sample=sample, X_fluid=X_fluid, **kwargs)
        return (4400 * XCO3) / (36.594 - 44*XCO3)  # Following Dixon 1997 setting Mr as constant

    def calculate_equilibrium_fluid_comp(self, pressure, sample, **kwargs):
        """ Returns 1.0 if a pure H2O fluid is saturated. Returns 0.0 if a pure H2O fluid is
        undersaturated.

        Parameters
        ----------
        pressure     float
            The total pressure of the system in bars.
        sample         Sample class
            Magma major element composition.

        Returns
        -------
        float
            1.0 if CO2-fluid saturated, 0.0 otherwise.
        """
        if self.calculate_saturation_pressure(sample=sample, **kwargs) < pressure:
            return 0.0
        else:
            return 1.0

    def calculate_saturation_pressure(self, sample, X_fluid=1.0, **kwargs):
        """
        Calculates the pressure at which a pure CO2 fluid is saturated, for the given sample
        composition and CO2 concentration. Calls the scipy.root_scalar routine, which makes
        repeated called to the calculate_dissolved_volatiles method.

        Parameters
        ----------
        sample         Sample class
            Magma major element composition (including CO2).
        X_fluid     float
            The mole fraction of CO2 in the fluid. Default is 1.0.

        Returns
        -------
        float
            Calculated saturation pressure in bars.
        """
        if isinstance(sample, sample_class.Sample) is False:
            raise core.InputError("Sample must be an instance of the Sample class.")
        if sample.check_oxide('CO2') is False:
            raise core.InputError("sample must contain CO2.")
        if sample.get_composition('CO2') < 0:
            raise core.InputError("Dissolved CO2 concentration must be greater than 0 wt%.")

        try:
            satP = root_scalar(
                self.root_saturation_pressure, x0=100.0, x1=1000.0, args=(sample, kwargs)).root
        except Exception:
            w.warn("Saturation pressure not found.", RuntimeWarning, stacklevel=2)
            satP = np.nan
        return np.real(satP)

    def molfrac_molecular(self, pressure, sample, X_fluid=1.0, **kwargs):
        """Calculates the mole fraction of CO3(-2) dissolved when in equilibrium with a pure CO2
        fluid at 1200C, using Eqn (1) of Dixon (1997).

        Parameters
        ----------
        pressure      float
            Total pressure in bars.
        sample        Sample class
            Magma major element composition.
        X_fluid     float
            Mole fraction of CO2 in the fluid.

        Returns
        -------
        float
            Mole fraction of CO3(2-) dissolved."""

        DeltaVr = 23  # Changed to match dixon spreadsheet.14 (cm3 mole-1)
        P0 = 1
        R = 83.15
        T0 = 1473.15

        fugacity = self.fugacity_model.fugacity(pressure=pressure, X_fluid=X_fluid, **kwargs)

        XCO3Std = self.XCO3_Std(sample)

        return XCO3Std * fugacity * np.exp(-DeltaVr * (pressure-P0)/(R*T0))

    def XCO3_Std(self, sample):
        """ Calculates the mole fraction of CO3(2-) dissolved when in equilibrium with pure CO2
        vapour at 1200C and 1 bar, using Eq (8) of Dixon (1997).

        Parameters
        ----------
        sample    Sample class
            Magma major element chemistry.

        Returns
        -------
        float
            Mole fraction of CO3(2-) dissolved at 1 bar and 1200C.
        """
        if sample.get_composition('SiO2') > 48.9:
            return 3.817e-7
        else:
            return 8.697e-6 - 1.697e-7*sample.get_composition('SiO2')

    def root_saturation_pressure(self, pressure, sample, kwargs):
        """ The function called by scipy.root_scalar when finding the saturation pressure using
        calculate_saturation_pressure.

        Parameters
        ----------
        pressure     float
            Total pressure in bars.
        sample         Sample class
            Magma major element composition.

        Returns
        -------
        float
            The difference between the dissolved CO2 the pressure guessed, and the CO2
            concentration passed in the sample variable.
        """
        return (self.calculate_dissolved_volatiles(pressure=pressure, sample=sample, **kwargs) -
                sample.get_composition('CO2'))


class water(model_classes.Model):
    """
    Implementation of the Dixon (1997) water solubility model, as a Model class.
    """

    def __init__(self):
        self.set_volatile_species(['H2O'])
        self.set_fugacity_model(fugacity_models.fugacity_MRK_h2o())
        self.set_activity_model(activity_models.activity_idealsolution())
        self.set_calibration_ranges([
            calibration_checks.CalibrationRange(
                'pressure', 1000, calibration_checks.crf_LessThan, 'bar',
                'Dixon (1997, Pi-SiO2 simpl.) Water',
                fail_msg=crmsg_1000bar_fail,
                pass_msg=calibration_checks.crmsg_LessThan_pass,
                description_msg=calibration_checks.crmsg_LessThan_description),
            calibration_checks.CalibrationRange(
                'pressure', 2000, calibration_checks.crf_LessThan, 'bar',
                'Dixon (1997, Pi-SiO2 simpl.) Water',
                fail_msg=crmsg_2000bar_fail,
                pass_msg=calibration_checks.crmsg_LessThan_pass,
                description_msg=calibration_checks.crmsg_LessThan_description),
            calibration_checks.CalibrationRange(
                'SiO2', 49, calibration_checks.crf_LessThan, 'wt%',
                'Dixon (1997, Pi-SiO2 simpl.) Water',
                fail_msg=crmsg_49_fail,
                pass_msg=calibration_checks.crmsg_LessThan_pass,
                description_msg=calibration_checks.crmsg_LessThan_description),
            calibration_checks.CalibrationRange(
                'SiO2', 40, calibration_checks.crf_GreaterThan, 'wt%',
                'Dixon (1997, Pi-SiO2 simpl.) Water',
                fail_msg=crmsg_40_fail,
                pass_msg=calibration_checks.crmsg_GreaterThan_pass,
                description_msg=calibration_checks.crmsg_GreaterThan_description),
            calibration_checks.CalibrationRange(
                'temperature', [1000, 1400], calibration_checks.crf_Between, 'oC',
                'Dixon (1997, Pi-SiO2 simpl.) Water',
                fail_msg=crmsg_BC_T,
                pass_msg=calibration_checks.crmsg_Between_pass,
                description_msg=calibration_checks.crmsg_Between_description)])
        self.set_solubility_dependence(False)

    def calculate_dissolved_volatiles(self, pressure, sample, X_fluid=1.0, **kwargs):
        """Calculates the dissolved H2O concentration using Eqns (5) and (6) of Dixon (1997).

        Parameters
        ----------
        pressure  float
            Total pressure in bars.
        sample      Sample class
            Magma major element composition.
        X_fluid      float
            The mol fraction of H2O in the fluid.

        Returns
        -------
        float
            The H2O concentration in wt%.
        """
        if isinstance(sample, sample_class.Sample) is False:
            raise core.InputError("Sample must be an instance of the Sample class.")
        if sample.check_oxide('SiO2') is False:
            raise core.InputError("sample must contain SiO2.")
        if pressure < 0:
            raise core.InputError("Pressure must be positive")
        if X_fluid < 0 or X_fluid > 1:
            raise core.InputError("X_fluid must have a value between 0 and 1.")

        if pressure == 0:
            return 0

        XH2O = self.molfrac_molecular(pressure=pressure, sample=sample, X_fluid=X_fluid, **kwargs)
        XOH = self.XOH(pressure=pressure, sample=sample, X_fluid=X_fluid, **kwargs)

        XB = XH2O + 0.5*XOH
        return 1801.5*XB/(36.594-18.579*XB)  # Following Dixon spreadsheet

    def calculate_equilibrium_fluid_comp(self, pressure, sample, **kwargs):
        """ Returns 1.0 if a pure H2O fluid is saturated.
        Returns 0.0 if a pure H2O fluid is undersaturated.

        Parameters
        ----------
        pressure     float
            The total pressure of the system in bars.
        sample       Sample class
            Magma major element composition (including H2O).

        Returns
        -------
        float
            1.0 if H2O-fluid saturated, 0.0 otherwise.
        """
        if self.calculate_saturation_pressure(sample=sample, **kwargs) < pressure:
            return 0.0
        else:
            return 1.0

    def calculate_saturation_pressure(self, sample, X_fluid=1.0, **kwargs):
        """
        Calculates the pressure at which a pure H2O fluid is saturated, for the given sample
        composition and H2O concentration. Calls the scipy.root_scalar routine, which makes
        repeated called to the calculate_dissolved_volatiles method.

        Parameters
        ----------
        sample      Sample class
            Magma major element composition (including H2O).
        X_fluid     float
            The mole fraction of H2O in the fluid. Default is 1.0.

        Returns
        -------
        float
            Calculated saturation pressure in bars.
        """
        if sample.check_oxide('H2O') is False:
            raise core.InputError("sample must contain H2O")
        if sample.get_composition('H2O') < 0:
            raise core.InputError("H2O concentration must be greater than 0 wt%.")
        try:
            satP = root_scalar(
                self.root_saturation_pressure, x0=100.0, x1=1000.0, args=(sample, kwargs)).root
        except Exception:
            w.warn("Saturation pressure not found.", RuntimeWarning, stacklevel=2)
            satP = np.nan
        return np.real(satP)

    def molfrac_molecular(self, pressure, sample, X_fluid=1.0, **kwargs):
        """Calculates the mole fraction of molecular H2O dissolved when in equilibrium with
        a pure H2O fluid at 1200C, using Eqn (2) of Dixon (1997).

        Parameters
        ----------
        pressure      float
            Total pressure in bars.
        sample      Sample class
            Magma major element composition.
        X_fluid     float
            Mole fraction of H2O in the fluid.

        Returns
        -------
        float
            Mole fraction of molecular H2O dissolved.
        """

        VH2O = 12  # cm3 mole-1
        P0 = 1
        R = 83.15
        T0 = 1473.15

        XH2OStd = self.XH2O_Std(sample)

        fugacity = self.fugacity_model.fugacity(pressure=pressure, X_fluid=X_fluid, **kwargs)

        return XH2OStd * fugacity * np.exp(-VH2O * (pressure-P0)/(R*T0))

    def XH2O_Std(self, sample):
        """ Calculates the mole fraction of molecular H2O dissolved when in equilibrium with pure
        H2O vapour at 1200C and 1 bar, using Eq (9) of Dixon (1997).

        Parameters
        ----------
        sample    Sample class
            Magma major element composition.

        Returns
        -------
        float
            Mole fraction of molecular water dissolved at 1 bar and 1200C.
        """
        if sample.get_composition('SiO2') > 48.9:
            return 3.28e-5
        else:
            return -3.04e-5 + 1.29e-6*sample.get_composition('SiO2')

    def XOH(self, pressure, sample, X_fluid=1.0, **kwargs):
        """
        Calculates the mole fraction of hydroxyl groups dissolved by solving Eq (4) of Dixon
        (1997). Calls scipy.root_scalar to find the root of the XOH_root method.

        Parameters
        ----------
        pressure     float
            Total pressure in bars.
        sample         pandas Series or dict
            Major element oxides in wt%.
        X_fluid     float
            Mole fraction of H2O in the fluid.

        Returns
        -------
        float
            Mole fraction of hydroxyl groups dissolved.
        """

        XH2O = self.molfrac_molecular(pressure=pressure, sample=sample, X_fluid=X_fluid, **kwargs)
        if XH2O < 1e-14:
            return 0
        return np.exp(root_scalar(self.XOH_root, x0=np.log(0.5), x1=np.log(0.1), args=(XH2O)).root)

    def XOH_root(self, XOH, XH2O):
        """
        Method called by scipy.root_scalar when finding the saturation pressure using the
        calculate_saturation_pressure method. Implements Eq (4) of Dixon (1997).

        Parameters
        ----------
        XOH         float
            Guess for the mole fraction of hydroxyl groups dissolved in melt.
        XH2O    float
            Mole fraction of molecular water dissolved in melt.

        Returns
        -------
        float
            The difference between the RHS and LHS of Eq (4) of Dixon (1997) for the
            guessed value of XOH.
        """

        A = 0.403
        B = 15.333
        C = 10.894

        XOH = np.exp(XOH)

        term = (XOH)**2.0/(XH2O*(1.0-XOH-XH2O))

        lhs = - np.log(term)

        rhs = A + B*XOH + C*XH2O

        return rhs - lhs

    def root_saturation_pressure(self, pressure, sample, kwargs):
        """ Function called by scipy.root_scalar when finding the saturation pressure using
        calculate_saturation_pressure.

        Parameters
        ----------
        pressure     float
            Pressure guess in bars
        sample         pandas Series or dict
            Major elements in wt% (normalized to 100%), including H2O.
        kwargs         dictionary
            Additional keyword arguments supplied to calculate_saturation_pressure. Might be
            required for the fugacity or activity models.

        Returns
        -------
        float
            The differece between the dissolved H2O at the pressure guessed, and the H2O
            concentration passed in the sample variable.
        """
        return (self.calculate_dissolved_volatiles(pressure=pressure, sample=sample, **kwargs) -
                sample.get_composition('H2O'))


crmsg_40_fail = ("{param_name} ({param_val:.1f} {units})<40 wt%, which is the lower calibration "
                 "limit of the Dixon (1997, Pi-SiO2 simpl.) Model. VESIcal has performed "
                 "calculations assuming SiO2=40wt%. ")
crmsg_1000bar_fail = ("{param_name} exceeds 1000 bar, which Iacono-Marziano et al. (2012) suggest"
                      " as an upper calibration limit of the Dixon (1997, Pi-SiO2 simpl.) Model, ")
crmsg_2000bar_fail = ("as well as the upper calibration limit of 2000 bar suggested by Lesne et"
                      " al. (2011), ")
crmsg_5000bar_fail = ("and the upper calibration limit of 5000 bar suggested by Newman and"
                      " Lowenstern, (2002). ")
crmsg_49_fail = ("{param_name} ({param_val:.1f} {units})>49 wt%, which is the calibration limit "
                 "of the Dixon (1997, Pi-SiO2 simpl.) Model. VESIcal has performed calculations "
                 "assuming SiO2=49wt% for this sample. ")
# Warning for Dixon temperature
crmsg_BC_T = ("{param_name} ({param_val:.1f} {units}) lies more than 200°C away from the "
              "temperature the Dixon (1997, Pi-SiO2 simpl.) model was calibrated for (the range "
              "suggested by Newman and Lowenstern, 2002; VolatileCalc). ")


mixed = model_classes.MixedFluid({'H2O': water(), 'CO2': carbon()})



# ---------- DEFINE SOME CONSTANTS ------------- #
oxides = ['SiO2', 'TiO2', 'Al2O3', 'Fe2O3', 'Cr2O3', 'FeO', 'MnO', 'MgO', 'NiO', 'CoO', 'CaO',
          'Na2O', 'K2O', 'P2O5', 'H2O', 'CO2']
anhydrous_oxides = ['SiO2', 'TiO2', 'Al2O3', 'Fe2O3', 'Cr2O3', 'FeO', 'MnO', 'MgO', 'NiO', 'CoO',
                    'CaO', 'Na2O', 'K2O', 'P2O5']
volatiles = ['H2O', 'CO2']
oxideMass = {'SiO2':  28.085 + 15.999 * 2,
             'MgO':   24.305 + 15.999,
             'FeO':   55.845 + 15.999,
             'CaO':   40.078 + 15.999,
             'Al2O3': 2 * 26.982 + 15.999 * 3,
             'Na2O':  22.99 * 2 + 15.999,
             'K2O':   39.098 * 2 + 15.999,
             'MnO':   54.938 + 15.999,
             'TiO2':  79.867,
             'P2O5':  2 * 30.974 + 5 * 15.999,
             'Cr2O3': 151.992,
             'NiO':   58.693 + 16,
             'CoO':   44.01,
             'Fe2O3': 55.845 * 2 + 15.999 * 3,
             'H2O':   18.02,
             'CO2':   44.01}

CationNum = {'SiO2': 1, 'MgO': 1, 'FeO': 1, 'CaO': 1, 'Al2O3': 2, 'Na2O': 2,
             'K2O': 2, 'MnO': 1, 'TiO2': 1, 'P2O5': 2, 'Cr2O3': 2,
             'NiO': 1, 'CoO': 1, 'Fe2O3': 2, 'H2O': 2, 'CO2': 1}

OxygenNum = {'SiO2': 2, 'MgO': 1, 'FeO': 1, 'CaO': 1, 'Al2O3': 3, 'Na2O': 1,
             'K2O': 1, 'MnO': 1, 'TiO2': 2, 'P2O5': 5, 'Cr2O3': 3,
             'NiO': 1, 'CoO': 1, 'Fe2O3': 3, 'H2O': 1, 'CO2': 2}

CationCharge = {'SiO2': 4, 'MgO': 2, 'FeO': 2, 'CaO': 2, 'Al2O3': 3, 'Na2O': 1,
                'K2O': 1, 'MnO': 2, 'TiO2': 4, 'P2O5': 5, 'Cr2O3': 3,
                'NiO': 2, 'CoO': 2, 'Fe2O3': 3, 'H2O': 1, 'CO2': 4}

CationMass = {'SiO2': 28.085, 'MgO': 24.305, 'FeO': 55.845, 'CaO': 40.078, 'Al2O3': 26.982,
              'Na2O': 22.990, 'K2O': 39.098, 'MnO': 54.938, 'TiO2': 47.867, 'P2O5': 30.974,
              'Cr2O3': 51.996, 'NiO': 58.693, 'CoO': 28.01, 'Fe2O3': 55.845, 'H2O': 1.01,
              'CO2': 12.011}

oxides_to_cations = {'SiO2': 'Si', 'MgO': 'Mg', 'FeO': 'Fe', 'CaO': 'Ca', 'Al2O3': 'Al',
                     'Na2O': 'Na', 'K2O': 'K', 'MnO': 'Mn', 'TiO2': 'Ti', 'P2O5': 'P',
                     'Cr2O3': 'Cr', 'NiO': 'Ni', 'CoO': 'Co', 'Fe2O3': 'Fe3', 'H2O': 'H',
                     'CO2': 'C', 'F': 'F'}
cations_to_oxides = {'Si': 'SiO2', 'Mg': 'MgO', 'Fe': 'FeO', 'Ca': 'CaO', 'Al': 'Al2O3',
                     'Na': 'Na2O', 'K': 'K2O', 'Mn': 'MnO', 'Ti': 'TiO2', 'P': 'P2O5',
                     'Cr': 'Cr2O3', 'Ni': 'NiO', 'Co': 'CoO', 'Fe3': 'Fe2O3', 'H': 'H2O',
                     'C': 'CO2'}


# ---------- DATA TRANSFORMATION FOR PANDAS DATAFRAMES --------- #
def fluid_molfrac_to_wt(data, H2O_colname='XH2O_fl_VESIcal', CO2_colname='XCO2_fl_VESIcal'):
    """
    Takes in a pandas dataframe object and converts only the fluid composition from mole fraction
    to wt%, leaving the melt composition in tact. The user must specify the names of the
    XH2O_fl and XCO2_fl columns.

    Parameters
    ----------
    data: pandas DataFrame
        Sample composition(s) containing columns for H2O and CO2 concentrations in the fluid.

    H2O_colname: str
        OPTIONAL. The default value is 'XH2O_fl', which is what is returned by BatchFile() core
        calculations. String containing the name of the column corresponding to the H2O
        concentration in the fluid, in mol fraction.

    CO2_colname: str
        OPTIONAL. The default value is 'XCO2_fl', which is what is returned by BatchFile() core
        calculations. String containing the name of the column corresponding to the CO2
        concentration in the fluid, in mol fraction.

    Returns
    -------
    pandas DataFrame
        Original data passed plus newly calculated values are returned.
    """
    convData = data.copy()

    MPO_H2O_list = []
    MPO_CO2_list = []
    for index, row in convData.iterrows():
        MPO_H2O_list.append(row[H2O_colname] * oxideMass["H2O"])
        MPO_CO2_list.append(row[CO2_colname] * oxideMass["CO2"])

    convData["MPO_H2O"] = MPO_H2O_list
    convData["MPO_CO2"] = MPO_CO2_list
    convData["H2O_fl_wt"] = 100 * convData["MPO_H2O"] / (convData["MPO_H2O"] + convData["MPO_CO2"])
    convData["CO2_fl_wt"] = 100 * convData["MPO_CO2"] / (convData["MPO_H2O"] + convData["MPO_CO2"])

    del convData["MPO_H2O"]
    del convData["MPO_CO2"]

    return convData


def fluid_wt_to_molfrac(data, H2O_colname='H2O_fl_wt', CO2_colname='CO2_fl_wt'):
    """
    Takes in a pandas dataframe object and converts only the fluid composition from wt% to mole
    fraction, leaving the melt composition in tact. The user must specify the names of the
    H2O_fl_wt and CO2_fl_wt columns.

    Parameters
    ----------
    data: pandas DataFrame
        DataFrame containing columns for H2O and CO2 concentrations in the fluid.

    H2O_colname: str
        OPTIONAL. The default value is 'H2O_fl_wt', which is what is returned by BatchFile() core
        calculations. String containing the name of the column corresponding to the H2O
        concentration in the fluid, in wt%.

    CO2_colname: str
        OPTIONAL. The default value is 'CO2_fl_wt', which is what is returned by BatchFile() core
        calculations. String containing the name of the column corresponding to the CO2
        concentration in the fluid, in wt%.

    Returns
    -------
    pandas DataFrame
        Original data passed plus newly calculated values are returned.
    """
    convData = data.copy()

    MPO_H2O_list = []
    MPO_CO2_list = []
    for index, row in convData.iterrows():
        MPO_H2O_list.append(row[H2O_colname] / oxideMass["H2O"])
        MPO_CO2_list.append(row[CO2_colname] / oxideMass["CO2"])

    convData["MPO_H2O"] = MPO_H2O_list
    convData["MPO_CO2"] = MPO_CO2_list
    convData["XH2O_fl"] = convData["MPO_H2O"] / (convData["MPO_H2O"] + convData["MPO_CO2"])
    convData["XCO2_fl"] = convData["MPO_CO2"] / (convData["MPO_H2O"] + convData["MPO_CO2"])

    del convData["MPO_H2O"]
    del convData["MPO_CO2"]

    return convData


class Error(Exception):
    """Base class for exceptions in this module."""
    pass


class InputError(Error):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class SaturationError(Error):
    """Exception raised for errors thrown when a sample does not reach saturation.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message
